import matlab
import numpy as np
import pytest
from numpy.testing import assert_allclose as assert_close

from at import linopt6


def _py_data(ml_data):
    """Convert a Matlab vector to a numpy vector"""
    return np.squeeze(np.asarray(ml_data))


def _ml_refs(ring, py_refs):
    """Convert refpoints to Matlab"""
    uintrefs = ring.uint32_refpts(py_refs)
    return matlab.double([float(ref + 1) for ref in uintrefs])


def _compare_4d(py_data, ml_data, fields, atol=1.0e-12, rtol=1.0e-7):
    for ml_key, py_key in fields:
        ml_val = _py_data(ml_data[ml_key])
        if py_key == "closed_orbit":
            py_val = np.squeeze(py_data[py_key][:, :4])
        else:
            py_val = np.squeeze(py_data[py_key])
        assert_close(py_val, ml_val, atol=atol, rtol=rtol)


def _compare_6d(py_data, ml_data, fields, atol=1.0e-12, rtol=1.0e-7):
    for ml_key, py_key in fields:
        ml_val = _py_data(ml_data[ml_key])
        py_val = np.squeeze(py_data[py_key])
        assert_close(py_val, ml_val, atol=atol, rtol=rtol)


@pytest.mark.parametrize("dp", (-0.01, 0.0, 0.01))
@pytest.mark.parametrize("lattices", ["dba", "hmba"])
def test_4d_analysis(engine, request, lattices, dp):
    """Compare linopt6 in 4D"""
    py_lattice, ml_lattice, _ = request.getfixturevalue(lattices)
    fields = [
        ("SPos", "s_pos"),
        ("ClosedOrbit", "closed_orbit"),
        ("M", "M"),
        ("A", "A"),
        ("Dispersion", "dispersion"),
        ("alpha", "alpha"),
        ("beta", "beta"),
        ("mu", "mu"),
    ]
    pypts = range(10)
    mlpts = _ml_refs(py_lattice, pypts)

    # python call
    py_data0, py_ringdata, py_data = linopt6(
        py_lattice, refpts=pypts, dp=dp, get_w=True
    )
    # Matlab call
    ml_ringdata, ml_data = engine.pyproxy(
        "atlinopt6", ml_lattice, mlpts, "dp", dp, "get_w", nargout=2
    )
    # Comparison
    assert_close(py_ringdata.tune, _py_data(ml_ringdata["tune"]), atol=1.0e-8, rtol=0)
    assert_close(
        py_ringdata.chromaticity,
        _py_data(ml_ringdata["chromaticity"]),
        atol=5.0e-4,
        rtol=0,
    )

    _compare_6d(py_data, ml_data, fields, atol=1.0e-6, rtol=1e-8)
    _compare_6d(py_data, ml_data, [("W", "W")], atol=1.0e-6, rtol=4.0e-4)


@pytest.mark.parametrize("lattices", ["hmba"])
def test_6d_analysis(engine, request, lattices):
    """Compare linopt6 in 6D"""
    py_lattice, ml_lattice, _ = request.getfixturevalue(lattices)
    fields = [
        ("SPos", "s_pos"),
        ("ClosedOrbit", "closed_orbit"),
        ("M", "M"),
        ("A", "A"),
        ("Dispersion", "dispersion"),
        ("alpha", "alpha"),
        ("beta", "beta"),
        ("mu", "mu"),
        ("W", "W"),
    ]
    ml_lattice = engine.pyproxy("atradon", ml_lattice)
    py_lattice = py_lattice.radiation_on(copy=True)
    pypts = range(10)
    mlpts = _ml_refs(py_lattice, pypts)

    # python call
    py_data0, py_ringdata, py_data = linopt6(py_lattice, refpts=pypts, get_w=True)
    # Matlab call
    ml_ringdata, ml_data = engine.pyproxy(
        "atlinopt6", ml_lattice, mlpts, "get_w", nargout=2
    )
    # Comparison
    assert_close(py_ringdata.tune, _py_data(ml_ringdata["tune"]), atol=1.0e-12, rtol=0)
    assert_close(
        py_ringdata.chromaticity,
        _py_data(ml_ringdata["chromaticity"]),
        atol=1.0e-5,
        rtol=0,
    )

    _compare_6d(py_data, ml_data, fields, atol=1.0e-6, rtol=1e-8)



"""Text representation of a python AT lattice with each element represented by
its :py:func:`repr` string
"""

from __future__ import annotations

__all__ = ["load_repr", "save_repr"]

import sys
from os.path import abspath
from typing import Optional

import numpy as np

# imports necessary in 'globals()' for 'eval'
from numpy import array, uint8, nan as NaN  # noqa: F401

from at.lattice import Lattice, Element

# imports necessary in' globals()' for 'eval'
from at.lattice import Particle  # noqa: F401
from at.load import register_format
from at.load.utils import element_classes, keep_attributes, keep_elements

# Map class names to Element classes
_CLASS_MAP = dict((cls.__name__, cls) for cls in element_classes())


def _element_from_string(elem_string: str) -> Element:
    """Builds an :py:class:`.Element` from its python :py:func:`repr` string

    Parameters:
        elem_string:    String representation of an :py:class:`.Element`

    Returns:
        elem (Element): new :py:class:`.Element`
    """
    return eval(elem_string, globals(), _CLASS_MAP)


def load_repr(filename: str, **kwargs) -> Lattice:
    """Create a :py:class:`.Lattice`  from a text repr-file

    Parameters:
        filename:           Name of a '.m' file

    Keyword Args:
        name (str):         Name of the lattice. Default: taken from
          the lattice
        energy (float):     Energy of the lattice [eV]. Default: taken
          from the lattice elements
        periodicity(int):   Number of periods. Default: taken from the
          elements, or 1
        *:                  All other keywords will be set as Lattice
          attributes

    Returns:
        lattice (Lattice):          New :py:class:`.Lattice` object

    See Also:
        :py:func:`.load_lattice` for a generic lattice-loading function.
    """

    def elem_iterator(params, repr_file):
        with open(params.setdefault("in_file", repr_file), "rt") as file:
            # the 1st line is the dictionary of saved lattice parameters
            for k, v in eval(next(file)).items():
                params.setdefault(k, v)
            for line in file:
                yield _element_from_string(line.strip())

    return Lattice(abspath(filename), iterator=elem_iterator, **kwargs)


def save_repr(ring: Lattice, filename: Optional[str] = None) -> None:
    """Save a :py:class:`.Lattice` as a repr-file

    Parameters:
        ring:           Lattice description
        filename:       Name of the '.repr' file. Default: outputs on
          :py:obj:`sys.stdout`

    See Also:
        :py:func:`.save_lattice` for a generic lattice-saving function.
    """

    def save(file):
        print(repr(keep_attributes(ring)), file=file)
        for elem in keep_elements(ring):
            print(repr(elem), file=file)

    # Set options to print the full representation of float variables
    with np.printoptions(formatter={"float_kind": repr}):
        if filename is None:
            save(sys.stdout)
        else:
            with open(filename, "wt") as reprfile:
                save(reprfile)


register_format(
    ".repr",
    load_repr,
    save_repr,
    descr=("Text representation of a python AT Lattice. "
           "See :py:func:`.load_repr`."),
)



"""Physical constants
"""

from math import sqrt as _sqrt, pi
# noinspection PyUnresolvedReferences
from scipy.constants import c as clight
from scipy.constants import physical_constants as _cst
from scipy.constants import e as qe

#: Electron mass [eV]
e_mass = 1.0e+06 * _cst['electron mass energy equivalent in MeV'][0]  # eV
#: Proton mass [eV]
p_mass = 1.0e+06 * _cst['proton mass energy equivalent in MeV'][0]    # eV

_e_radius = _cst['classical electron radius'][0]
_hbar_c = _cst['reduced Planck constant times c in MeV fm'][0]

#: :math:`C_\gamma` [m/eV^3]
Cgamma = 4.0 * pi * _e_radius / 3.0 / pow(e_mass, 3)                 # m/eV^3
#: :math:`C_q` [m]
Cq = 55 / 32 / _sqrt(3) * _hbar_c / e_mass * 1.0e-9                  # m



import numpy
import pytest

from at.lattice import checktype, get_cells, refpts_iterator, get_elements
from at.lattice import elements, uint32_refpts, bool_refpts, checkattr
from at.lattice import get_s_pos, tilt_elem, shift_elem, set_tilt, set_shift


@pytest.mark.parametrize(
    "ref_in, expected",
    (
        [2, numpy.array([2], dtype=numpy.uint32)],
        [-1, numpy.array([4], dtype=numpy.uint32)],
        [7, numpy.array([2], dtype=numpy.uint32)],
        [[1, 2, 3], numpy.array([1, 2, 3], dtype=numpy.uint32)],
        [[0, 1, 2, 3, 4, 5], numpy.array([0, 1, 2, 3, 4, 5], dtype=numpy.uint32)],
        [[0, 6, 2, -2, 4, 5], numpy.array([0, 1, 2, 3, 4, 5], dtype=numpy.uint32)],
    ),
)
def test_uint32_refpts_converts_numerical_inputs_correctly(ref_in, expected):
    numpy.testing.assert_equal(uint32_refpts(ref_in, 5), expected)
    # float indexes are deprecated
    # ref_in2 = numpy.asarray(ref_in).astype(numpy.float64)
    # numpy.testing.assert_equal(uint32_refpts(ref_in2, 5), expected)
    ref_in2 = numpy.asarray(ref_in).astype(numpy.int64)
    numpy.testing.assert_equal(uint32_refpts(ref_in2, 5), expected)


@pytest.mark.parametrize(
    "ref_in, expected",
    (
        [None, numpy.array([], dtype=numpy.uint32)],
        [[], numpy.array([], dtype=numpy.uint32)],
        [True, numpy.array([0], dtype=numpy.uint32)],
        [False, numpy.array([], dtype=numpy.uint32)],
        [[True, False, True], numpy.array([0, 2], dtype=numpy.uint32)],
        [
            [False, False, False, False, False, False],
            numpy.array([], dtype=numpy.uint32),
        ],
        [
            [True, True, True, True, True, True],
            numpy.array([0, 1, 2, 3, 4, 5], dtype=numpy.uint32),
        ],
    ),
)
def test_uint32_refpts_converts_other_input_types_correctly(ref_in, expected):
    numpy.testing.assert_equal(uint32_refpts(ref_in, 5), expected)


# too long, misordered, duplicate, -ve indexing misordered, -ve indexing
# duplicate, wrap-around indexing misordered, wrap-around indexing duplicate,
# -ve indexing and wrap-around indexing duplicate, -ve indexing and wrap-around
# indexing misordered
@pytest.mark.parametrize(
    "ref_in",
    ([0, 1, 2, 3], [2, 1], [0, 0], [-1, 0], [0, -2], [3, 0], [1, 3], [-1, 3], [3, -2]),
)
def test_uint32_refpts_throws_IndexError_correctly(ref_in):
    with pytest.raises(IndexError):
        uint32_refpts(ref_in, 2)


@pytest.mark.parametrize(
    "ref_in, expected",
    (
        [[True], numpy.array([True, False, False, False], dtype=bool)],
        [[], numpy.array([False, False, False, False], dtype=bool)],
        [None, numpy.array([False, False, False, False], dtype=bool)],
        [2, numpy.array([False, False, True, False], dtype=bool)],
        [[1, 2], numpy.array([False, True, True, False], dtype=bool)],
    ),
)
def test_bool_refpts(ref_in, expected):
    numpy.testing.assert_equal(bool_refpts(ref_in, 3), expected)


@pytest.mark.parametrize("ref_in", (10, range(100)))
def test_bool_refpts_throws_IndexErrorrs(ref_in):
    with pytest.raises(IndexError):
        bool_refpts(ref_in, 3)


def test_checkattr(simple_ring):
    assert checkattr("Length")(simple_ring[0]) is True
    assert checkattr("not_an_attr")(simple_ring[0]) is False
    assert list(filter(checkattr("Length", 1), simple_ring)) == [
        simple_ring[0],
        simple_ring[3],
        simple_ring[4],
        simple_ring[5],
    ]
    assert list(filter(checkattr("Length", 2), simple_ring)) == []
    assert list(filter(checkattr("not_an_attr"), simple_ring)) == []


def test_checktype(simple_ring):
    assert checktype(elements.Drift)(simple_ring[0]) is True
    assert checktype(elements.Marker)(simple_ring[0]) is False
    assert list(filter(checktype(elements.Drift), simple_ring)) == [
        simple_ring[0],
        simple_ring[3],
        simple_ring[4],
        simple_ring[5],
    ]
    assert list(filter(checktype(elements.Monitor), simple_ring)) == []


def test_get_cells(simple_ring):
    a = numpy.array([True, True, True, True, True, True, False])
    numpy.testing.assert_equal(get_cells(simple_ring, checkattr("Length")), a)
    a = numpy.array([False, True, False, False, False, False, False])
    numpy.testing.assert_equal(get_cells(simple_ring, "attr"), a)
    a = numpy.array([True, False, False, False, False, False, False])
    numpy.testing.assert_equal(get_cells(simple_ring, "FamName", "D1"), a)


def test_refpts_iterator(simple_ring):
    assert list(refpts_iterator(simple_ring, [0, 1, 2, 3, 4, 5])) == simple_ring
    assert list(refpts_iterator(simple_ring, numpy.ones(6, dtype=bool))) == simple_ring
    assert list(refpts_iterator(simple_ring, [1])) == [simple_ring[1]]
    a = numpy.array([False, True, False, False, False, False])
    assert list(refpts_iterator(simple_ring, a)) == [simple_ring[1]]


def test_get_elements(hmba_lattice):
    # test FamName direct match
    assert get_elements(hmba_lattice, "BPM_06") == [hmba_lattice[65]]
    # test FamName wildcard matching
    assert get_elements(hmba_lattice, "QD2?") == list(hmba_lattice[9, 113])
    assert get_elements(hmba_lattice, "QD3*") == list(hmba_lattice[19, 105])
    assert get_elements(hmba_lattice, "S*H2B") == list([hmba_lattice[55]])
    assert get_elements(hmba_lattice, "*C_1") == list(hmba_lattice[59, 60])
    assert get_elements(hmba_lattice, "DR_2[1-3]") == list(hmba_lattice[54, 56, 58])
    assert get_elements(hmba_lattice, "DR_2[!1-7]") == list(hmba_lattice[52, 78, 80])
    # test element instance
    marker = elements.Marker("M1")
    assert get_elements(hmba_lattice, marker) == list(hmba_lattice[1, 12, 61, 67, 73])
    # test element type
    assert get_elements(hmba_lattice, elements.RFCavity) == [hmba_lattice[0]]
    # test invalid key raises TypeError
    with pytest.raises(TypeError):
        get_elements(hmba_lattice, 1.0)


def test_get_s_pos_returns_zero_for_empty_lattice():
    numpy.testing.assert_equal(get_s_pos([]), numpy.array((0,)))


def test_get_s_pos_returns_length_for_lattice_with_one_element():
    e = elements.LongElement("e", 0.1)
    assert get_s_pos([e], [1]) == numpy.array([0.1])


def test_get_s_pos_returns_all_pts_for_lat_with_2_elements():
    e = elements.LongElement("e", 0.1)
    f = elements.LongElement("e", 2.1)
    numpy.testing.assert_equal(get_s_pos([e, f]), numpy.array([0, 0.1, 2.2]))


def test_get_s_pos_returns_all_pts_for_lat_with_2_elements_using_int_refpts():
    e = elements.LongElement("e", 0.1)
    f = elements.LongElement("e", 2.1)
    lat = [e, f]
    numpy.testing.assert_equal(
        get_s_pos(lat, range(len(lat) + 1)), numpy.array([0, 0.1, 2.2])
    )


def test_get_s_pos_returns_all_pts_for_lat_with_2_elements_using_bool_refpts():
    e = elements.LongElement("e", 0.1)
    f = elements.LongElement("e", 2.1)
    lat = [e, f]
    numpy.testing.assert_equal(
        get_s_pos(lat, numpy.ones(3, dtype=bool)), numpy.array([0, 0.1, 2.2])
    )


def test_tilt_elem(simple_ring):
    tilt_elem(simple_ring[0], (numpy.pi / 4))
    v = 1 / 2**0.5
    a = numpy.diag([v, v, v, v, 1.0, 1.0])
    a[0, 2], a[1, 3], a[2, 0], a[3, 1] = v, v, -v, -v
    numpy.testing.assert_allclose(simple_ring[0].R1, a)
    numpy.testing.assert_allclose(simple_ring[0].R2, a.T)


def test_shift_elem(simple_ring):
    shift_elem(simple_ring[2], 1.0, 0.5)
    a = numpy.array([1.0, 0.0, 0.5, 0.0, 0.0, 0.0])
    numpy.testing.assert_equal(simple_ring[2].T1, -a)
    numpy.testing.assert_equal(simple_ring[2].T2, a)


def test_set_tilt(simple_ring):
    set_tilt(simple_ring, [(numpy.pi / 4), 0, 0, 0, (numpy.pi / 4), 0])
    v = 1 / 2**0.5
    a = numpy.diag([v, v, v, v, 1.0, 1.0])
    a[0, 2], a[1, 3], a[2, 0], a[3, 1] = v, v, -v, -v
    numpy.testing.assert_allclose(simple_ring[0].R1, a)
    numpy.testing.assert_allclose(simple_ring[0].R2, a.T)
    numpy.testing.assert_allclose(simple_ring[0].R1, a)
    numpy.testing.assert_allclose(simple_ring[0].R2, a.T)
    ring = [simple_ring[0]]
    set_tilt(ring, 0)
    numpy.testing.assert_allclose(ring[0].R1, numpy.eye(6))
    numpy.testing.assert_allclose(ring[0].R2, numpy.eye(6))


def test_set_shift(simple_ring):
    set_shift(
        simple_ring,
        numpy.array([0.0, 0.0, 0.0, 1.0, 0.0, 0.5]),
        numpy.array([0.0, 0.0, 0.0, 2.0, 0.0, 1.0]),
    )
    a = numpy.array([0.5, 0.0, 1.0, 0.0, 0.0, 0.0])
    numpy.testing.assert_equal(simple_ring[3].T1, -a * 2)
    numpy.testing.assert_equal(simple_ring[3].T2, a * 2)
    numpy.testing.assert_equal(simple_ring[5].T1, -a)
    numpy.testing.assert_equal(simple_ring[5].T2, a)
    ring = [simple_ring[3]]
    set_shift(ring, 3, 5)
    a = numpy.array([3.0, 0.0, 5.0, 0.0, 0.0, 0.0])
    numpy.testing.assert_equal(simple_ring[3].T1, -a)
    numpy.testing.assert_equal(simple_ring[3].T2, a)



import numpy
from at import radiation_parameters
from at.constants import clight, qe
from scipy.interpolate import interp1d
import time
from at.collective import Wake
from at.lattice import Lattice


class Haissinski(object):
    r"""
    Class to find the longitudinal distribution in the presence
    of a short range wakefield.

    Parameters: 
    
        wake_object:    class initialized with Wake that contains a 
          longitudinal wake Z and equivalent srange.
        ring:           a lattice object that is used to extract 
         machine parameters

    Keyword Args:
        m (int):        the number of points in the full
          distribution
        kmax:           the min and max of the range of the distribution.
          See equation 27. In units of [:math:`\sigma_z0`]
        current:        bunch current.
        numIters (int): the number of iterations
        eps:            the convergence criteria.

    This class is a direct implementation of the following paper:
    "Numerical solution of the Haïssinski equation for the equilibrium
    state of  a stored electron beam", R. Warnock, K.Bane, Phys. Rev. Acc.
    and Beams 21, 124401 (2018)

    The reader is referred to this paper for a description of the methods.
    The equation number of key formula are written next to the relevant
    function.

    The functions solve or solve_steps can be used after initialisation
    An example usage can be found in:

    at/pyat/examples/Collective/LongDistribution.py

    Future developments of this class:
        Adding LR wake or harmonic cavity as done at SOLEIL. Needs
        to be added WITH this class which is just for short range wake.
    """

    def __init__(self, wake_object: Wake, ring: Lattice, m: int = 12, kmax: float = 1,
                 current: float = 1e-4, numIters: int = 10, eps: float = 1e-10):

        self.circumference = ring.circumference
        self.energy = ring.energy

        radpars = radiation_parameters(ring)
        self.f_s = radpars.f_s
        self.nu_s = self.f_s / (clight/self.circumference)
        self.sigma_e = radpars.sigma_e * self.energy
        self.sigma_l = radpars.sigma_l

        #  The paper uses alpha and eta with same sign.
        #  AT uses opposite signs. So negative sign is needed for eta.
        self.eta = -ring.radiation_off(copy=True).slip_factor
        self.ga = ring.gamma
        self.betrel = ring.beta

        self.numIters = numIters
        self.eps = eps

        #  negative s to be consistent with paper and negative Wz
        s = wake_object.srange/self.sigma_l
        self.ds = numpy.diff(s)[-1]
        self.s = -s
        self.wtot_fun = interp1d(self.s, -wake_object.Z,
                                 bounds_error=False, fill_value = 0)

        if m % 2 != 0:
            raise AttributeError('m must be even and int')
        self.m = m
        self.npoints = 2*self.m + 1

        self.kmax = kmax
        self.dq = self.kmax/self.m

        self.q_array = -self.kmax + numpy.arange(self.npoints)*self.dq
        self.set_weights()
        self.dtFlag = False

        self.set_I(current)
        self.initial_phi()

        try:
            self.Sfun
        except AttributeError:
            self.precompute_S()

        self.compute_Smat()

    def set_weights(self):
        '''
        Page 7 second paragraph, in the text
        '''
        self.weights = numpy.ones(self.npoints)*2
        self.weights[1::2] = self.weights[1::2]**2
        self.weights[0] = 1
        self.weights[-1] = 1
        self.weights *= self.dq/3

    def precompute_S(self):
        '''
        Equation 16
        '''
        print('Computing integrated wake potential')
        sr = numpy.arange(2*numpy.amin(self.q_array),
                          numpy.abs(2*numpy.amin(self.q_array))
                          + self.ds, self.ds)
        res = numpy.zeros(len(sr))

        topend = numpy.trapz(self.wtot_fun(numpy.arange(numpy.amax(sr),
                             numpy.amax(self.s), self.ds)),
                             x=numpy.arange(numpy.amax(sr),
                             numpy.amax(self.s), self.ds))
        for p, low in enumerate(sr):
            if p % 10000 == 0:
                print(p, ' out of ', len(sr))
            rr = numpy.arange(low, numpy.amax(sr), self.ds)
            val = numpy.trapz(self.wtot_fun(rr), x=rr) + topend
            res[p] = val
        #  Not used except for plotting
        self.Sfun_range = sr
        self.Sfun = interp1d(sr, res)

    def compute_Smat(self):
        '''
        The sampling of the integrated wake potential S
        is only made at certain places. So all possibilities
        are loaded into a matrix for speed.
        '''
        self.Smat = numpy.zeros((self.npoints, self.npoints))
        for iq1, q1 in enumerate(self.q_array):
            for iq2, q2 in enumerate(self.q_array):
                self.Smat[iq1, iq2] = self.Sfun(q1-q2)

    def set_I(self, current):
        '''
        Equation 11
        '''
        self.N = current*self.circumference/(clight*self.betrel*qe)
        self.Ic = numpy.sign(self.eta)*qe*self.N/(2*numpy.pi *
                                                  self.nu_s*self.sigma_e)
        #  removed one qe for eV to joule
        print('Normalised current: ', self.Ic*1e12, ' pC/V')

    def initial_phi(self):
        '''
        Simply a gaussian but using the normalised units.
        Page 5 top right, in the text.
        '''
        self.phi = numpy.exp(-self.q_array**2/2)*self.Ic/numpy.sqrt(2*numpy.pi)
        self.phi_0 = self.phi.copy()

    def Fi(self):
        '''
        Equation 28
        '''
        self.allFi = numpy.zeros(self.npoints)
        for i in numpy.arange(self.npoints):
            sum1 = 0
            for j in numpy.arange(self.npoints):
                sum1b = 0
                for k in numpy.arange(self.npoints):
                    sum1b += self.weights[k] * self.Smat[j, k] * self.phi[k]
                sum1 += self.weights[j] * \
                    numpy.exp(-self.q_array[j]**2/2 + sum1b)
            sum2 = 0
            for j in numpy.arange(self.npoints):
                sum2 += self.weights[j] * self.Smat[i, j] * self.phi[j]
            self.allFi[i] = self.phi[i] * sum1 - self.Ic * \
                numpy.exp(-self.q_array[i]**2/2 + sum2)

    def dFi_ij(self, i, j):

        sum1 = 0
        for k in numpy.arange(self.npoints):

            sum1b = 0
            for li in numpy.arange(self.npoints):

                sum1b += self.weights[li] * self.Smat[k, li] * self.phi[li]

            #  in paper i starts from 1, but kron(0,0) is 0 in python.
            #  No good for python, so i need this to be able to loop from 0
            kron = 0 if i != j else 1

            sum1 += self.weights[k] * \
                (kron + self.phi[i] * self.weights[j] * self.Smat[k, j]) * \
                numpy.exp(-self.q_array[k]**2/2 + sum1b)

        sum2 = 0
        for k in numpy.arange(self.npoints):
            sum2 += self.weights[k] * self.Smat[i, k] * self.phi[k]

        val = sum1 - self.Ic*self.weights[j] * self.Smat[i, j] \
            * numpy.exp(-self.q_array[i]**2/2 + sum2)

        return val

    def dFi_dphij(self):
        '''
        Equation 30
        '''
        if not self.dtFlag:
            t0 = time.time()
            self.dFi_ij(0, 0)
            dt = time.time() - t0
            print('Computing dFi/dphij, will take approximately ',
                  numpy.round(dt*self.npoints**2/60, 3),
                  ' minutes per iteration')
            self.dtFlag = True
        allDat = [self.dFi_ij(i, j) for i in numpy.arange(self.npoints)
                  for j in numpy.arange(self.npoints)]
        self.alldFi_dphij = numpy.reshape(allDat, (self.npoints, self.npoints))

    def compute_new_phi(self):
        ainv = numpy.linalg.inv(self.alldFi_dphij)
        self.pseudo_inv = numpy.dot(ainv, -self.allFi)
        self.phi_1 = self.pseudo_inv + self.phi

    def update(self):
        self.phi = self.phi_1.copy()

    def convergence(self):
        '''
        Equation 32
        '''
        self.conv = (numpy.sum(numpy.abs(self.phi_1 - self.phi))) \
            / numpy.sum(numpy.abs(self.phi))

    def set_output(self):
        self.res = (self.phi_1.copy())[::-1]

    def solve(self):
        for it in numpy.arange(self.numIters):
            self.Fi()
            self.dFi_dphij()
            self.compute_new_phi()
            self.convergence()
            self.set_output()
            print('Iteration: ', it, ', Delta: ', self.conv)
            if self.conv < self.eps:
                print('Converged')
                break
            if it != self.numIters-1:
                self.update()
        if self.conv > self.eps:
            print('Did not converge, maybe solve for an intermediate '
                  'current then use the solution as a starting point')

    def solve_steps(self, currents):
        '''
        INPUT:
            currents  an array of currents to solve. If 0 is given,
                      a current of 10uA is used to prevent failure.
        '''
        self.I_steps = numpy.zeros(len(currents))
        self.res_steps = numpy.zeros((len(currents), len(self.q_array)))
        for ii, Ib in enumerate(currents):
            print('Running step ', ii+1, ' out of ', len(currents))
            #  If Ib = 0, the gaussian is zero. Small epsilon is given
            if Ib == 0:
                self.set_I(1e-5)
            else:
                self.set_I(Ib)
            self.I_steps[ii] = self.Ic
            self.solve()
            self.res_steps[ii, :] = self.res



"""Lattice synoptics"""
# noinspection PyPackageRequirements
import matplotlib.axes
import numpy
# noinspection PyPackageRequirements
import matplotlib.pyplot as plt
# noinspection PyPackageRequirements
from matplotlib.patches import Polygon
# noinspection PyPackageRequirements
from matplotlib.collections import PatchCollection
from at.lattice import Lattice, Refpts, elements as elts

__all__ = ['plot_synopt']

# Default properties for element representation
DIPOLE = dict(label='Dipoles', facecolor=(0.5, 0.5, 1.0))
QUADRUPOLE = dict(label='Quadrupoles', facecolor=(1.0, 0.5, 0.5))
SEXTUPOLE = dict(label='Sextupoles', facecolor=(0.5, 1.0, 0.5))
MULTIPOLE = dict(label='Multipoles', facecolor=(0.25, 0.75, 0.25))
MONITOR = dict(label='Monitors', linestyle=None, marker=10, color='k')


# noinspection PyDefaultArgument
def plot_synopt(ring: Lattice, axes: matplotlib.axes.Axes = None,
                dipole: dict = {}, quadrupole: dict = {}, sextupole: dict = {},
                multipole: dict = {}, monitor: dict = {},
                labels: Refpts = None):
    """Plots a synoptic of a lattice

    Parameters:
        ring:           Lattice description.
        axes:           :py:class:`~matplotlib.axes.Axes` for plotting the
          synoptic. If :py:obj:`None`, a new figure will be created. Otherwise,
          a new axes object sharing the same x-axis as the given one is created.
        labels:         display the name of selected elements.
        dipole:         Dictionary of properties overloading the default
          properties. If :py:obj:`None`, dipoles will not be shown.
        quadrupole:     Same definition as for dipole
        sextupole:      Same definition as for dipole
        multipole:      Same definition as for dipole
        monitor:        Same definition as for dipole

    Returns:
        synopt_axes (Axes): Synoptic axes
     """

    class Dipole(Polygon):
        xx = numpy.array([0, 0, 1, 1], dtype=float)
        yy = numpy.array([0, 1, 1, 0], dtype=float)

        def __init__(self, s, length, **kwargs):
            xy = numpy.stack((self.xx * length + s, self.yy), axis=1)
            super(Dipole, self).__init__(xy, closed=False, **kwargs)

    class Quadrupole(Polygon):
        xx = numpy.array([0, 0, 0.5, 1, 1])
        yy = {True: numpy.array([0, 1, 1.4, 1, 0]),
              False: numpy.array([0, 1, 0.6, 1, 0])}

        def __init__(self, s, length, foc, **kwargs):
            xy = numpy.stack((self.xx * length + s, self.yy[foc]), axis=1)
            super(Quadrupole, self).__init__(xy, closed=False, **kwargs)

    class Sextupole(Polygon):
        xx = numpy.array([0, 0, 0.33, 0.66, 1, 1])
        yy = {True: numpy.array([0, 0.8, 1, 1, 0.8, 0]),
              False: numpy.array([0, 0.8, 0.6, 0.6, 0.8, 0])}

        def __init__(self, s, length, foc, **kwargs):
            xy = numpy.stack((self.xx * length + s, self.yy[foc]), axis=1)
            super(Sextupole, self).__init__(xy, closed=False, **kwargs)

    class Multipole(Polygon):
        xx = numpy.array([0, 0, 1, 1], dtype=float)
        yy = numpy.array([0, 0.8, 0.8, 0])

        def __init__(self, s, length, **kwargs):
            xy = numpy.stack((self.xx * length + s, self.yy), axis=1)
            super(Multipole, self).__init__(xy, closed=False, **kwargs)

    class Monitor(Polygon):
        xx = numpy.array([0.0, 0.0])
        yy = numpy.array([0.0, 1.2])

        def __init__(self, s, **kwargs):
            xy = numpy.stack((self.xx + s, self.yy), axis=1)
            super(Monitor, self).__init__(xy, closed=False, **kwargs)

    def ismultipole(elem):
        return isinstance(elem, elts.Multipole) and not isinstance(elem, (
            elts.Dipole, elts.Quadrupole, elts.Sextupole))

    if axes is None:
        fig = plt.figure()
        axsyn = fig.add_subplot(111, xlim=ring.s_range)
    else:
        axsyn = axes.twinx()
    axsyn.set_axis_off()         # Set axis invisible
    axsyn.set_ylim((0.0, 20.0))  # Initial scaling of elements
    axsyn.set_zorder(-0.2)       # Put synoptic in the background

    s_pos = ring.get_s_pos(range(len(ring)))

    if dipole is not None:
        props = DIPOLE.copy()
        props.update(dipole)
        dipoles = PatchCollection(
            (Dipole(s, el.Length) for s, el in zip(s_pos, ring)
             if isinstance(el, elts.Dipole)), **props)
        axsyn.add_collection(dipoles)

    if quadrupole is not None:
        props = QUADRUPOLE.copy()
        props.update(quadrupole)
        quadrupoles = PatchCollection(
            (Quadrupole(s, el.Length, el.PolynomB[1] >= 0.0)
             for s, el in zip(s_pos, ring)
             if isinstance(el, elts.Quadrupole)), **props)
        axsyn.add_collection(quadrupoles)

    if sextupole is not None:
        props = SEXTUPOLE.copy()
        props.update(sextupole)
        sextupoles = PatchCollection(
            (Sextupole(s, el.Length, el.PolynomB[2] >= 0.0)
             for s, el in zip(s_pos, ring)
             if isinstance(el, elts.Sextupole)), **props)
        axsyn.add_collection(sextupoles)

    if multipole is not None:
        props = MULTIPOLE.copy()
        props.update(multipole)
        multipoles = PatchCollection(
            (Multipole(s, el.Length) for s, el in zip(s_pos, ring)
             if ismultipole(el)), **props)
        axsyn.add_collection(multipoles)

    if monitor is not None:
        props = MONITOR.copy()
        props.update(monitor)
        s = s_pos[[isinstance(el, elts.Monitor) for el in ring]]
        y = numpy.zeros(s.shape)
        # noinspection PyUnusedLocal
        monitors = axsyn.plot(s, y, **props)

    for idx in ring.get_uint32_index(labels):
        el = ring[idx]
        s = s_pos[idx]
        axsyn.text(s + 0.5*el.Length, 1.6, el.FamName[:10],
                   rotation='vertical', horizontalalignment='center',
                   fontsize='small')

    return axsyn



from .elements import Dipole
import numpy as np
from scipy.optimize import fsolve
from math import sin, cos

__all__ = []


def rbendtune(self: Dipole) -> None:
    # noinspection PyUnresolvedReferences
    r"""Set *X0ref* and *RefDZ* for rectangular bending magnets

        This method must be called after creating a rectangular bending magnet
        or after setting its *PolynomA/B* attributes. It will set the correct *X0ref*
        and *RefDZ* attributes to get a zero closed orbit for the reference particle.

        The method will do nothing on dipoles with a non-rectangular passmethod.

        Example:

            >>> # Identify the rectangular bends
            >>> rbends = ring.get_bool_index(...)
            >>> # Set their correct attributes
            >>> for dip in ring.select(rbends):
            ...     dip.rbendtune()

        """

    def cross(x0r: float):
        """Return the horizontal exit angle of the reference particle"""
        elem.X0ref = x0r
        out = elem.track(np.zeros(6))
        return out[1]

    def checkmul(el):
        """Check if there are multipole"""
        for order in range(el.MaxOrder + 1):
            if el.PolynomB[order] != 0.0:
                return True
        return False

    passmethod = self.PassMethod.replace("RadPass", "Pass")
    if any(
        [
            passmethod == pm
            for pm in (
                "BndStrMPoleSymplectic4Pass",
                "ExactRectangularBendPass",
                "ExactRectBendPass",
            )
        ]
    ):
        elem = self.copy()
        elem.PassMethod = passmethod
        theta = elem.BendingAngle

        # Analytical estimate
        x0ref = elem.Length * ((cos(0.5 * theta) - 1.0) / theta + sin(0.5 * theta) / 12)

        # Search if there are multipoles
        if checkmul(self):
            x0ref = float(fsolve(cross, x0ref))

        self.X0ref = x0ref
        rout = self.track(np.zeros(6))
        self.RefDZ = rout[5]


Dipole.rbendtune = rbendtune



import at
import numpy
from numpy.testing import assert_allclose


def test_1d_acceptance(hmba_lattice):
    hmba_lattice = hmba_lattice.radiation_off(copy=True)
    acceptance, _, _ = hmba_lattice.get_horizontal_acceptance(1e-3, 1.0e-3)
    expected = numpy.array([-0.011, 0.011])
    assert_allclose(acceptance, expected, atol=1e-6)


def test_1d_acceptance_refpts(hmba_lattice):
    hmba_lattice = hmba_lattice.radiation_off(copy=True)
    acceptance, _, _ = hmba_lattice.get_horizontal_acceptance(1e-3, 1.0e-3,
                                                              refpts=10)
    expected = numpy.array([-0.002, 0.002])
    assert_allclose(acceptance, expected, atol=1e-6)


def test_2d_acceptance(hmba_lattice):
    hmba_lattice = hmba_lattice.radiation_off(copy=True)
    acceptance, _, _ = hmba_lattice.get_acceptance(['x', 'y'], [10, 5],
                                                   [10.0e-3, 10.0e-3],
                                                   grid_mode=at.GridMode.RECURSIVE)
    expected = numpy.array([[-0.01125, -0.0053033, 0., 0.00574524, 0.010625],
                            [0., 0.0053033, 0.006875, 0.00574524, 0.]])
    assert_allclose(acceptance, expected, atol=1e-6, rtol=1.0e-4)



"""
Coupled or non-coupled 4x4 linear motion
"""
from __future__ import annotations
import numpy
from math import sqrt, pi, sin, cos, atan2
from collections.abc import Callable
import warnings
from scipy.linalg import solve
from ..constants import clight
from ..lattice import DConstant, Refpts, get_bool_index, get_uint32_index
from ..lattice import AtWarning, Lattice, Orbit, check_6d, get_s_pos
from ..lattice import AtError
from ..lattice import frequency_control
from ..tracking import internal_lpass
from .orbit import find_orbit4, find_orbit6
from .matrix import find_m44, find_m66
from .amat import a_matrix, jmat, jmatswap
from .harmonic_analysis import get_tunes_harmonic

__all__ = ['linopt', 'linopt2', 'linopt4', 'linopt6', 'avlinopt',
           'get_optics', 'get_tune', 'get_chrom']

_S = jmat(1)
_S2 = numpy.array([[0, 1], [-1, 0]], dtype=numpy.float64)

# dtype for structured array containing linopt parameters
_DATA2_DTYPE = [('alpha', numpy.float64, (2,)),
                ('beta', numpy.float64, (2,)),
                ('mu', numpy.float64, (2,))
                ]

_DATA4_DTYPE = [('alpha', numpy.float64, (2,)),
                ('beta', numpy.float64, (2,)),
                ('mu', numpy.float64, (2,)),
                ('gamma', numpy.float64),
                ('A', numpy.float64, (2, 2)),
                ('B', numpy.float64, (2, 2)),
                ('C', numpy.float64, (2, 2))
                ]

_DATA6_DTYPE = [('alpha', numpy.float64, (2,)),
                ('beta', numpy.float64, (2,)),
                ('mu', numpy.float64, (3,)),
                ('R', numpy.float64, (3, 6, 6)),
                ('A', numpy.float64, (6, 6)),
                ('dispersion', numpy.float64, (4,)),
                ]

_DATAX_DTYPE = [('alpha', numpy.float64, (2,)),
                ('beta', numpy.float64, (2,)),
                ('mu', numpy.float64, (2,)),
                ('R', numpy.float64, (2, 4, 4)),
                ('A', numpy.float64, (4, 4)),
                ]

_W24_DTYPE = [('W', numpy.float64, (2,)),
              ('Wp', numpy.float64, (2,)),
              ('dalpha', numpy.float64, (2,)),
              ('dbeta', numpy.float64, (2,)),
              ('dmu', numpy.float64, (2,)),
              ('ddispersion', numpy.float64, (4,)),
              ]


_W6_DTYPE = [('W', numpy.float64, (2,)),
             ('Wp', numpy.float64, (2,)),
             ('dalpha', numpy.float64, (2,)),
             ('dbeta', numpy.float64, (2,)),
             ('dmu', numpy.float64, (3,)),
             ('dR', numpy.float64, (3, 6, 6)),
             ('ddispersion', numpy.float64, (4,)),
             ]

_WX_DTYPE = [('W', numpy.float64, (2,)),
             ('Wp', numpy.float64, (2,)),
             ('dalpha', numpy.float64, (2,)),
             ('dbeta', numpy.float64, (2,)),
             ('dmu', numpy.float64, (2,)),
             ('dR', numpy.float64, (2, 4, 4)),
             ('ddispersion', numpy.float64, (4,)),
             ]


_IDX_DTYPE = [('idx', numpy.uint32)]


def _twiss22(t12, alpha0, beta0):
    """Propagate Twiss parameters"""
    bbb = t12[:, 0, 1]
    aaa = t12[:, 0, 0] * beta0 - bbb * alpha0
    beta = (aaa * aaa + bbb * bbb) / beta0
    alpha = -(aaa * (t12[:, 1, 0] * beta0 - t12[:, 1, 1] * alpha0) +
              bbb * t12[:, 1, 1]) / beta0
    mu = numpy.arctan2(bbb, aaa)
    # Unwrap negative jumps in betatron phase advance
    # dmu = numpy.diff(numpy.append([0], mu))
    # jumps = dmu < -1.0e-3
    # mu += numpy.cumsum(jumps) * 2.0 * numpy.pi
    return alpha, beta, mu


def _closure(m22):
    diff = (m22[0, 0] - m22[1, 1]) / 2.0
    try:
        sinmu = numpy.sign(m22[0, 1]) * sqrt(-m22[0, 1]*m22[1, 0] - diff*diff)
        cosmu = 0.5 * numpy.trace(m22)
        alpha = diff / sinmu
        beta = m22[0, 1] / sinmu
        return alpha, beta, cosmu + sinmu*1j
    except ValueError:          # Unstable motion
        return numpy.nan, numpy.nan, numpy.nan


# noinspection PyShadowingNames,PyPep8Naming
def _tunes(ring, **kwargs):
    """"""
    if ring.is_6d:
        nd = 3
        mt, _ = find_m66(ring, **kwargs)
    else:
        nd = 2
        mt, _ = find_m44(ring, **kwargs)
    try:
        _, vps = a_matrix(mt)
        tunes = numpy.mod(numpy.angle(vps) / 2.0 / pi, 1.0)
    except AtError:
        warnings.warn(AtWarning('Unstable ring'))
        tunes = numpy.empty(nd)
        tunes[:] = numpy.nan
    return tunes


def _analyze2(mt, ms):
    """Analysis of a 2D 1-turn transfer matrix"""
    A = mt[:2, :2]
    B = mt[2:, 2:]
    alp0_a, bet0_a, vp_a = _closure(A)
    alp0_b, bet0_b, vp_b = _closure(B)
    vps = numpy.array([vp_a, vp_b])
    el0 = (numpy.array([alp0_a, alp0_b]),
           numpy.array([bet0_a, bet0_b]),
           0.0)
    alpha_a, beta_a, mu_a = _twiss22(ms[:, :2, :2], alp0_a, bet0_a)
    alpha_b, beta_b, mu_b = _twiss22(ms[:, 2:, 2:], alp0_b, bet0_b)
    els = (numpy.stack((alpha_a, alpha_b), axis=1),
           numpy.stack((beta_a, beta_b), axis=1),
           numpy.stack((mu_a, mu_b), axis=1))
    return vps, _DATA2_DTYPE, el0, els, _W24_DTYPE


def _analyze4(mt, ms):
    """Analysis of a 4D 1-turn transfer matrix according to Sagan, Rubin"""
    def propagate(t12):
        M = t12[:2, :2]
        N = t12[2:, 2:]
        m = t12[:2, 2:]
        n = t12[2:, :2]
        ff = n @ C + g * N
        gamma = sqrt(numpy.linalg.det(ff))
        e12 = (g * M - m @ _S @ C.T @ _S.T) / gamma
        f12 = ff / gamma
        a12 = e12 @ A @ _S @ e12.T @ _S.T
        b12 = f12 @ B @ _S @ f12.T @ _S.T
        c12 = M @ C + g * m @ _S @ f12.T @ _S.T
        return e12, f12, gamma, a12, b12, c12

    M = mt[:2, :2]
    N = mt[2:, 2:]
    m = mt[:2, 2:]
    n = mt[2:, :2]
    H = m + _S @ n.T @ _S.T
    detH = numpy.linalg.det(H)
    if detH == 0.0:
        g = 1.0
        C = -H
        A = M
        B = N
    else:
        t = numpy.trace(M - N)
        t2 = t * t
        t2h = t2 + 4.0 * detH
        g2 = (1.0 + sqrt(t2 / t2h)) / 2
        g = sqrt(g2)
        C = -H * numpy.sign(t) / (g * sqrt(t2h))
        A = g2 * M - g * (m @ _S @ C.T @ _S.T + C @ n) + \
            C @ N @ _S @ C.T @ _S.T
        B = g2 * N + g * (_S @ C.T @ _S.T @ m + n @ C) + \
            _S @ C.T @ _S.T @ M @ C
    alp0_a, bet0_a, vp_a = _closure(A)
    alp0_b, bet0_b, vp_b = _closure(B)
    vps = numpy.array([vp_a, vp_b])
    el0 = (numpy.array([alp0_a, alp0_b]),
           numpy.array([bet0_a, bet0_b]),
           0.0, g, A, B, C)
    if ms.shape[0] > 0:
        e, f, g, ai, bi, ci = zip(*[propagate(mi) for mi in ms])
        alp_a, bet_a, mu_a = _twiss22(numpy.array(e), alp0_a, bet0_a)
        alp_b, bet_b, mu_b = _twiss22(numpy.array(f), alp0_b, bet0_b)
        els = (numpy.stack((alp_a, alp_b), axis=1),
               numpy.stack((bet_a, bet_b), axis=1),
               numpy.stack((mu_a, mu_b), axis=1), numpy.array(g),
               numpy.stack(ai, axis=0), numpy.stack(bi, axis=0),
               numpy.stack(ci, axis=0))
    else:
        els = (numpy.empty((0, 2)), numpy.empty((0, 2)),
               numpy.empty((0, 2)), numpy.empty((0,)),
               numpy.empty((0, 2, 2)), numpy.empty((0, 2, 2)),
               numpy.empty((0, 2, 2)))
    return vps, _DATA4_DTYPE, el0, els, _W24_DTYPE


def _analyze6(mt, ms):
    """Analysis of a 2D, 4D, 6D 1-turn transfer matrix
    according to Wolski"""
    def get_phase(a22):
        """Return the phase for A standardization"""
        return atan2(a22[0, 1], a22[0, 0])

    def standardize(aa, slcs):
        """Apply rotation to put A in std form"""
        def rot2(slc):
            rot = -get_phase(aa[slc, slc])
            cs = cos(rot)
            sn = sin(rot)
            return aa[:, slc] @ numpy.array([[cs, sn], [-sn, cs]])

        return numpy.concatenate([rot2(slc) for slc in slcs], axis=1)

    def r_matrices(ai):
        # Rk = A * S * Ik * inv(A) * S.T
        def mul2(slc):
            return ai[:, slc] @ tt[slc, slc]

        ais = numpy.concatenate([mul2(slc) for slc in slices], axis=1)
        invai = solve(ai, ss.T)
        ri = numpy.array(
            [ais[:, sl] @ invai[sl, :] for sl in slices])
        mui = numpy.array([get_phase(ai[sl, sl]) for sl in slices])
        return mui, ri, ai

    def propagate4(ri, phi, ai):
        betai = numpy.stack((ri[..., 0, 0, 0], ri[..., 1, 2, 2]), axis=-1)
        alphai = -numpy.stack((ri[..., 0, 1, 0], ri[..., 1, 3, 2]), axis=-1)
        return alphai, betai, phi, ri, ai

    def propagate6(ri, phi, ai):
        alphai, betai, phi, ri, ai = propagate4(ri, phi, ai)
        dispersion = ri[..., 2, :4, 4] / ri[..., 2, 4, 4, numpy.newaxis]
        return alphai, betai, phi, ri, ai, dispersion

    nv = mt.shape[0]
    dms = nv // 2
    slices = [slice(2 * i, 2 * (i + 1)) for i in range(dms)]
    ss = jmat(dms)
    tt = jmatswap(dms)
    if dms >= 3:
        propagate = propagate6
        dtype = _DATA6_DTYPE
        wtype = _W6_DTYPE
    else:
        propagate = propagate4
        dtype = _DATAX_DTYPE
        wtype = _WX_DTYPE
    a0, vps = a_matrix(mt)

    astd = standardize(a0, slices)
    phi0, r0, _ = r_matrices(astd)
    el0 = propagate(r0, phi0, astd)
    if ms.shape[0] > 0:
        ps, rs, aas = zip(*[r_matrices(mi @ astd) for mi in ms])
        els = propagate(numpy.array(rs), numpy.array(ps), numpy.array(aas))
    elif dms >= 3:
        els = (numpy.empty((0, dms)), numpy.empty((0, dms)),
               numpy.empty((0, dms)),
               numpy.empty((0, dms, nv, nv)), numpy.empty((0, nv, nv)),
               numpy.empty((0, 4)))
    else:
        els = (numpy.empty((0, dms)), numpy.empty((0, dms)),
               numpy.empty((0, dms)),
               numpy.empty((0, dms, nv, nv)), numpy.empty((0, nv, nv)))
    return vps, dtype, el0, els, wtype


# noinspection PyShadowingNames,PyPep8Naming
def _linopt(ring: Lattice, analyze, refpts=None, dp=None, dct=None, df=None,
            orbit=None, twiss_in=None, get_chrom=False, get_w=False,
            keep_lattice=False, mname='M', add0=(), adds=(), cavpts=None,
            **kwargs):
    """"""
    def build_sigma(orbit, dp=None):
        """Build the initial distribution at entrance of the transfer line"""
        try:
            d0 = twiss_in['dispersion']
        except (ValueError, KeyError):  # record arrays throw ValueError !
            d0 = numpy.zeros((4,))

        try:
            rmat = twiss_in['R']
        except (ValueError, KeyError):  # record arrays throw ValueError !
            rmat = None

        try:
            alphas = twiss_in['alpha']
            betas = twiss_in['beta']
        except (ValueError, KeyError):  # record arrays throw ValueError !
            alphas = numpy.zeros((2,))
            betas = numpy.ones((2,))

        dorbit = numpy.hstack((d0, 1.0, 0.0))
        if orbit is None:
            try:
                orbit = twiss_in['closed_orbit']
            except (ValueError, KeyError):  # record arrays throw ValueError !
                orbit = numpy.zeros((6,))
            if dp is not None:
                orbit = orbit + dorbit * dp

        if dp is not None:
            try:
                drmat = twiss_in['dR']
            except (ValueError, KeyError):  # record arrays throw ValueError !
                drmat = None

            try:
                dd0 = twiss_in['ddispersion']
                dalpha = twiss_in['dalpha']
                dbeta = twiss_in['dbeta']
            except (ValueError, KeyError):  # record arrays throw ValueError !
                msg = ("'get_w' option for a line requires 'twiss_in' calculated "
                       "with 'get_w' activated")
                raise AtError(msg)

            orbit = orbit + numpy.hstack((dd0, 1.0, 0.0)) * dp * dp
            dorbit = numpy.hstack((d0+dd0*dp, 1.0, 0.0))

            if (rmat is not None) and (drmat is not None):
                rmat = rmat + drmat * dp
            else:
                alphas = alphas + dalpha * dp
                betas = betas + dbeta * dp

        if rmat is not None:
            # For some reason, "emittances" must be different...
            sigm = rmat[0, ...]+10.0*rmat[1, ...]
            if rmat.shape[0] >= 3:
                sigm = sigm+0.1*rmat[2, ...]
                dorbit = rmat[2, :, 4] / rmat[2, 4, 4, numpy.newaxis]
        else:
            slices = [slice(2 * i, 2 * (i + 1)) for i in range(2)]
            ab = numpy.stack((alphas, betas), axis=1)
            sigm = numpy.zeros((4, 4))
            for slc, (alpha, beta) in zip(slices, ab):
                gamma = (1.0+alpha*alpha)/beta
                sigm[slc, slc] = numpy.array([[beta, -alpha], [-alpha, gamma]])

        return orbit, sigm, dorbit

    def chrom_w(ringup, ringdn, orbitup, orbitdn,
                refpts=None, **kwargs):
        """Compute the chromaticity and W-functions"""
        # noinspection PyShadowingNames
        def off_momentum(rng, orb0, dp=None, **kwargs):
            if twiss_in is None:
                mt, ms = get_matrix(rng, refpts=refpts, orbit=orb0, **kwargs)
                mxx = mt
                dpup = orb0[4] + 0.5 * dp_step
                dpdn = orb0[4] - 0.5 * dp_step
                o0up = None
                o0dn = None
            else:
                orbit, sigma, dorbit = build_sigma(orb0, dp=dp)
                mt, ms = get_matrix(ring, refpts=refpts, orbit=orbit, **kwargs)
                mxx = sigma @ jmat(sigma.shape[0] // 2)
                o0up = orbit + dorbit * 0.5 * dp_step
                o0dn = orbit - dorbit * 0.5 * dp_step
                dpup = None
                dpdn = None
            vps, _, el0, els, wtype = analyze(mxx, ms)
            tunes = _tunes(rng, orbit=orb0)
            o0up, oup = get_orbit(ring, refpts=refpts, guess=orb0, dp=dpup,
                                  orbit=o0up, **kwargs)
            o0dn, odn = get_orbit(ring, refpts=refpts, guess=orb0, dp=dpdn,
                                  orbit=o0dn, **kwargs)
            d0 = (o0up - o0dn)[:4] / dp_step
            ds = numpy.array([(up - dn)[:4] / dp_step for up, dn in zip(oup, odn)])
            return tunes, el0, els, d0, ds, wtype

        def wget(ddp, elup, eldn, has_r):
            """Compute the chromatic amplitude function"""
            *data_up, = elup  # Extract alpha and beta
            *data_dn, = eldn
            alpha_up, beta_up, mu_up = data_up[:3]
            alpha_dn, beta_dn, mu_dn = data_dn[:3]
            db = numpy.array(beta_up - beta_dn) / ddp
            mb = (beta_up + beta_dn) / 2
            da = numpy.array(alpha_up - alpha_dn) / ddp
            ma = (alpha_up + alpha_dn) / 2
            wa = da - ma / mb * db
            wb = db / mb
            ww = numpy.sqrt(wa ** 2 + wb ** 2)
            wp = numpy.arctan2(wa, wb)
            dmu = numpy.array(mu_up - mu_dn) / ddp
            data_out = (ww, wp, da, db, dmu)
            if has_r:
                r_up = data_up[3]
                r_dn = data_dn[3]
                data_out += (numpy.array(r_up - r_dn) / ddp, )
            return data_out

        deltap = orbitup[4] - orbitdn[4]
        *data_up, = off_momentum(ringup, orbitup, dp=0.5*deltap,
                                 **kwargs)
        *data_dn, = off_momentum(ringdn, orbitdn, dp=-0.5*deltap,
                                 **kwargs)
        tunesup, el0up, elsup, d0up, dsup, wtype = data_up
        tunesdn, el0dn, elsdn, d0dn, dsdn, _ = data_dn
        has_r = len(wtype) == 7
        # in 6D, dp comes out of find_orbit6
        chrom = (tunesup-tunesdn) / deltap
        dd0 = numpy.array(d0up - d0dn) / deltap
        dds = numpy.array(dsup - dsdn) / deltap
        data0 = wget(deltap, el0up, el0dn, has_r)
        datas = wget(deltap, elsup, elsdn, has_r)
        data0 = data0 + (dd0,)
        datas = datas + (dds,)
        return chrom, data0, datas

    def unwrap(mu):
        """Remove the phase jumps"""
        dmu = numpy.diff(numpy.concatenate((numpy.zeros((1, dms)),
                                            mu)), axis=0)
        jumps = dmu < -1.e-3
        mu += numpy.cumsum(jumps, axis=0) * 2.0 * numpy.pi

    dp_step = kwargs.get('DPStep', DConstant.DPStep)
    addtype = kwargs.pop('addtype', [])

    if ring.is_6d:
        get_matrix = find_m66
        get_orbit = find_orbit6
    else:
        get_matrix = find_m44
        get_orbit = find_orbit4

    o0up = None
    o0dn = None
    if twiss_in is None:   # Ring
        if orbit is None:
            orbit, _ = get_orbit(ring, dp=dp, dct=dct, df=df,
                                 keep_lattice=keep_lattice, **kwargs)
            keep_lattice = True
        # Get 1-turn transfer matrix
        mt, ms = get_matrix(ring, refpts=refpts, orbit=orbit, **kwargs)
        mxx = mt
    else:                       # Transfer line
        orbit, sigma, dorbit = build_sigma(orbit)
        # Get 1-turn transfer matrix
        mt, ms = get_matrix(ring, refpts=refpts, orbit=orbit, **kwargs)
        mxx = sigma @ jmat(sigma.shape[0] // 2)
        o0up = orbit + dorbit * 0.5 * dp_step
        o0dn = orbit - dorbit * 0.5 * dp_step

    # Perform analysis
    vps, dtype, el0, els, wtype = analyze(mxx, ms)
    tunes = _tunes(ring, orbit=orbit)

    if (get_chrom or get_w) and mt.shape == (6, 6):
        f0 = ring.get_rf_frequency(cavpts=cavpts)
        df = dp_step * ring.disable_6d(copy=True).slip_factor * f0
        rgup = ring.set_rf_frequency(f0 + 0.5 * df, cavpts=cavpts, copy=True)
        rgdn = ring.set_rf_frequency(f0 - 0.5 * df, cavpts=cavpts, copy=True)
        if o0up is None:
            o0up, _ = get_orbit(rgup, guess=orbit, orbit=o0up, **kwargs)
            o0dn, _ = get_orbit(rgdn, guess=orbit, orbit=o0dn, **kwargs)
    else:
        rgup = ring
        rgdn = ring

    # Propagate the closed orbit
    orb0, orbs = get_orbit(ring, refpts=refpts, orbit=orbit,
                           keep_lattice=keep_lattice)
    spos = ring.get_s_pos(refpts)

    nrefs = orbs.shape[0]
    dms = vps.size
    if dms >= 3:            # 6D processing
        dtype = dtype + [('closed_orbit', numpy.float64, (6,)),
                         ('M', numpy.float64, (2*dms, 2*dms)),
                         ('s_pos', numpy.float64)]
        data0 = (orb0, numpy.identity(2*dms), 0.0)
        datas = (orbs, ms, spos)
        length = ring.get_s_pos(len(ring))[0]
        damping_rates = -numpy.log(numpy.absolute(vps))
        damping_times = length / clight / damping_rates
    else:               # 4D processing
        kwargs['keep_lattice'] = True
        dpup = orb0[4] + 0.5 * dp_step
        dpdn = orb0[4] - 0.5 * dp_step
        o0up, oup = get_orbit(ring, refpts=refpts, guess=orb0, dp=dpup,
                              orbit=o0up, **kwargs)
        o0dn, odn = get_orbit(ring, refpts=refpts, guess=orb0, dp=dpdn,
                              orbit=o0dn, **kwargs)
        d0 = (o0up - o0dn)[:4] / dp_step
        ds = numpy.array([(up - dn)[:4] / dp_step for up, dn in zip(oup, odn)])
        dtype = dtype + [('dispersion', numpy.float64, (4,)),
                         ('closed_orbit', numpy.float64, (6,)),
                         (mname, numpy.float64, (2*dms, 2*dms)),
                         ('s_pos', numpy.float64)]
        data0 = (d0, orb0, mt, get_s_pos(ring, len(ring))[0])
        datas = (ds, orbs, ms, spos)
        damping_times = numpy.nan

    if get_w:
        dtype = dtype + wtype
        chrom, ddata0, ddatas = chrom_w(rgup, rgdn, o0up, o0dn,
                                        refpts, **kwargs)
        data0 = data0 + ddata0
        datas = datas + ddatas
    elif get_chrom:
        tunesup = _tunes(rgup, orbit=o0up)
        tunesdn = _tunes(rgdn, orbit=o0dn)
        deltap = o0up[4] - o0dn[4]
        chrom = (tunesup - tunesdn) / deltap
    else:
        chrom = numpy.nan

    beamdata = numpy.array((tunes, chrom, damping_times),
                           dtype=[('tune', numpy.float64, (dms,)),
                                  ('chromaticity', numpy.float64, (dms,)),
                                  ('damping_time', numpy.float64, (dms,))
                                  ]).view(numpy.recarray)

    dtype = dtype + addtype
    elemdata0 = numpy.array(el0+data0+add0, dtype=dtype).view(numpy.recarray)
    elemdata = numpy.recarray((nrefs,), dtype=dtype)
    if nrefs > 0:
        for name, value in zip(numpy.dtype(dtype).names, els+datas+adds):
            elemdata[name] = value
        unwrap(elemdata.mu)
    return elemdata0, beamdata, elemdata


@check_6d(False)
def linopt2(ring: Lattice, *args, **kwargs):
    r"""Linear analysis of an uncoupled lattice

    Parameters:
        ring:   Lattice description.

    Keyword Args:
        refpts (Refpts):        Elements at which data is returned.
          It can be:

          1. an integer in the range [-len(ring), len(ring)-1]
             selecting the element according to python indexing rules.
             As a special case, len(ring) is allowed and refers to the end
             of the last element,
          2. an ordered list of such integers without duplicates,
          3. a numpy array of booleans of maximum length len(ring)+1,
             where selected elements are :py:obj:`True`.
        dp (float):             Momentum deviation. Defaults to :py:obj:`None`
        dct (float):            Path lengthening. Defaults to :py:obj:`None`
        df (float):             Deviation of RF frequency. Defaults to
          :py:obj:`None`
        orbit (Orbit):          Avoids looking for the closed orbit if it is
          already known ((6,) array)
        get_chrom (bool):       Compute chromaticities. Needs computing
          the tune at 2 different momentum deviations around the central one.
        get_w (bool):           Computes chromatic amplitude functions
          (W, WP) [4]_, and derivatives of the dispersion and twiss parameters
          versus dp. Needs to compute the optics at 2 different momentum
          deviations around the central one.
        keep_lattice (bool):    Assume no lattice change since the
          previous tracking. Defaults to :py:obj:`False`
        XYStep (float):         Step size.
          Default: :py:data:`DConstant.XYStep <.DConstant>`
        DPStep (float):         Momentum step size.
          Default: :py:data:`DConstant.DPStep <.DConstant>`
        twiss_in:               Initial conditions for transfer line optics.
          Record array as output by :py:func:`.linopt6`, or dictionary. Keys:

          R or alpha, beta
            mandatory (2,) arrays
          closed_orbit
            Optional (6,) array, default 0
          dispersion
            Optional (6,) array, default 0

          If present, the attribute **R** will be used, otherwise the
          attributes **alpha** and **beta** will be used. All other attributes
          are ignored.

    Returns:
        elemdata0:      Linear optics data at the entrance of the ring
        ringdata:       Lattice properties
        elemdata:       Linear optics at the points refered to by *refpts*,
          if refpts is :py:obj:`None` an empty lindata structure is returned.

    **elemdata** is a record array with fields:

    ================    ===================================================
    **s_pos**           longitudinal position [m]
    **M**               (4, 4) transfer matrix M from the beginning of ring
                        to the entrance of the element [2]_
    **closed_orbit**    (6,) closed orbit vector
    **dispersion**      (4,) dispersion vector
    **beta**            :math:`\left[ \beta_x,\beta_y \right]` vector
    **alpha**           :math:`\left[ \alpha_x,\alpha_y \right]` vector
    **mu**              :math:`\left[ \mu_x,\mu_y \right]`, betatron phase
                        (modulo :math:`2\pi`)
    **W**               :math:`\left[ W_x,W_y \right]` only if *get_w*
                        is :py:obj:`True`: chromatic amplitude function
    **Wp**              :math:`\left[ Wp_x,Wp_y \right]` only if *get_w*
                        is :py:obj:`True`: chromatic phase function
    **dalpha**          (2,) alpha derivative vector
                        (:math:`\Delta \alpha/ \delta_p`)
    **dbeta**           (2,) beta derivative vector
                        (:math:`\Delta \beta/ \delta_p`)
    **dmu**             (2,) mu derivative vector
                        (:math:`\Delta \mu/ \delta_p`)
    **ddispersion**     (4,) dispersion derivative vector
                        (:math:`\Delta D/ \delta_p`)
    ================    ===================================================

    All values given at the entrance of each element specified in refpts.
    Field values can be obtained with either *lindata['idx']* or *lindata.idx*

    **ringdata** is a record array with fields:

    =================   ======
    **tune**            Fractional tunes
    **chromaticity**    Chromaticities, only computed if *get_chrom* is
                        :py:obj:`True`
    =================   ======

    References:
        **[1]** D.Edwards,L.Teng IEEE Trans.Nucl.Sci. NS-20, No.3, p.885-888
        , 1973

        .. [2] E.Courant, H.Snyder

        **[3]** D.Sagan, D.Rubin Phys.Rev.Spec.Top.-Accelerators and beams,
        vol.2 (1999)

        .. [4] Brian W. Montague Report LEP Note 165, CERN, 1979
    """
    return _linopt(ring, _analyze2, *args, **kwargs)


@check_6d(False)
def linopt4(ring: Lattice, *args, **kwargs):
    r"""Linear analysis of a H/V coupled lattice

    4D-analysis of coupled motion following Sagan/Rubin

    Parameters:
        ring:   Lattice description.

    Keyword Args:
        refpts (Refpts):        Elements at which data is returned.
          It can be:

          1. an integer in the range [-len(ring), len(ring)-1]
             selecting the element according to python indexing rules.
             As a special case, len(ring) is allowed and refers to the end
             of the last element,
          2. an ordered list of such integers without duplicates,
          3. a numpy array of booleans of maximum length len(ring)+1,
             where selected elements are :py:obj:`True`.
        dp (float):             Momentum deviation. Defaults to :py:obj:`None`
        dct (float):            Path lengthening. Defaults to :py:obj:`None`
        df (float):             Deviation of RF frequency. Defaults to
          :py:obj:`None`
        orbit (Orbit):          Avoids looking for the closed orbit if it is
          already known ((6,) array)
        get_chrom (bool):       Compute chromaticities. Needs computing
          the tune at 2 different momentum deviations around the central one.
        get_w (bool):           Computes chromatic amplitude functions
          (W) [8]_. Needs to compute the optics at 2 different momentum
          deviations around the central one.
        keep_lattice (bool):    Assume no lattice change since the
          previous tracking. Defaults to :py:obj:`False`
        XYStep (float):         Step size.
          Default: :py:data:`DConstant.XYStep <.DConstant>`
        DPStep (float):         Momentum step size.
          Default: :py:data:`DConstant.DPStep <.DConstant>`
        twiss_in:               Initial conditions for transfer line optics.
          Record array as output by :py:func:`.linopt6`, or dictionary. Keys:

          R or alpha, beta
            mandatory (2,) arrays
          closed_orbit
            Optional (6,) array, default 0
          dispersion
            Optional (6,) array, default 0

          If present, the attribute **R** will be used, otherwise the
          attributes **alpha** and **beta** will be used. All other attributes
          are ignored.

    Returns:
        elemdata0:      Linear optics data at the entrance of the ring
        ringdata:       Lattice properties
        elemdata:       Linear optics at the points refered to by *refpts*,
          if refpts is :py:obj:`None` an empty lindata structure is returned.

    **elemdata** is a record array with fields:

    ================    ===================================================
    **s_pos**           longitudinal position [m]
    **M**               (4, 4) transfer matrix M from the beginning of ring
                        to the entrance of the element [6]_
    **closed_orbit**    (6,) closed orbit vector
    **dispersion**      (4,) dispersion vector
    **beta**            :math:`\left[ \beta_x,\beta_y \right]` vector
    **alpha**           :math:`\left[ \alpha_x,\alpha_y \right]` vector
    **mu**              :math:`\left[ \mu_x,\mu_y \right]`, betatron phase
                        (modulo :math:`2\pi`)
    **gamma**           gamma parameter of the transformation to
                        eigenmodes [7]_
    **W**               :math:`\left[ W_x,W_y \right]` only if *get_w*
                        is :py:obj:`True`: chromatic amplitude function
    **Wp**              :math:`\left[ Wp_x,Wp_y \right]` only if *get_w*
                        is :py:obj:`True`: chromatic phase function
    **dalpha**          (2,) alpha derivative vector
                        (:math:`\Delta \alpha/ \delta_p`)
    **dbeta**           (2,) beta derivative vector
                        (:math:`\Delta \beta/ \delta_p`)
    **dmu**             (2,) mu derivative vector
                        (:math:`\Delta \mu/ \delta_p`)
    **ddispersion**     (4,) dispersion derivative vector
                        (:math:`\Delta D/ \delta_p`)
    ================    ===================================================

    All values given at the entrance of each element specified in refpts.
    Field values can be obtained with either
    *lindata['idx']* or *lindata.idx*

    **ringdata** is a record array with fields:

    =================   ======
    **tune**            Fractional tunes
    **chromaticity**    Chromaticities, only computed if *get_chrom* is
                        :py:obj:`True`
    =================   ======

    References:
        **[5]** D.Edwards,L.Teng IEEE Trans.Nucl.Sci. NS-20, No.3, p.885-888
        , 1973

        .. [6] E.Courant, H.Snyder

        .. [7] D.Sagan, D.Rubin Phys.Rev.Spec.Top.-Accelerators and beams,
           vol.2 (1999)

        .. [8] Brian W. Montague Report LEP Note 165, CERN, 1979
    """
    return _linopt(ring, _analyze4, *args, **kwargs)


@frequency_control
def linopt6(ring: Lattice, *args, **kwargs):
    r"""Linear analysis of a fully coupled lattice using normal modes

    For circular machines, :py:func:`linopt6` analyses

    * the 4x4 1-turn transfer matrix if *ring* is 4D, or
    * the 6x6 1-turn transfer matrix if  *ring* is 6D.

    For a transfer line, The "twiss_in" intput must contain either:

    *  a field **R**, as provided by ATLINOPT6, or
    * the fields **beta** and **alpha**, as provided by linopt and linopt6

    Parameters:
        ring:   Lattice description.

    Keyword Args:
        refpts (Refpts):        Elements at which data is returned.
          It can be:

          1. an integer in the range [-len(ring), len(ring)-1]
             selecting the element according to python indexing rules.
             As a special case, len(ring) is allowed and refers to the end
             of the last element,
          2. an ordered list of such integers without duplicates,
          3. a numpy array of booleans of maximum length len(ring)+1,
             where selected elements are :py:obj:`True`.
        dp (float):             Momentum deviation. Defaults to :py:obj:`None`
        dct (float):            Path lengthening. Defaults to :py:obj:`None`
        df (float):             Deviation of RF frequency. Defaults to
          :py:obj:`None`
        orbit (Orbit):          Avoids looking for the closed orbit if it is
          already known ((6,) array)
        get_chrom (bool):       Compute chromaticities. Needs computing
          the tune at 2 different momentum deviations around the central one.
        get_w (bool):           Computes chromatic amplitude functions
          (W) [11]_. Needs to compute the optics at 2 different momentum
          deviations around the central one.
        keep_lattice (bool):    Assume no lattice change since the
          previous tracking. Defaults to :py:obj:`False`
        XYStep (float):         Step size.
          Default: :py:data:`DConstant.XYStep <.DConstant>`
        DPStep (float):         Momentum step size.
          Default: :py:data:`DConstant.DPStep <.DConstant>`
        twiss_in:               Initial conditions for transfer line optics.
          Record array as output by :py:func:`.linopt6`, or dictionary. Keys:

          R or alpha, beta
            mandatory (2,) arrays
          closed_orbit
            Optional (6,) array, default 0
          dispersion
            Optional (6,) array, default 0

          If present, the attribute **R** will be used, otherwise the
          attributes **alpha** and **beta** will be used. All other attributes
          are ignored.
        cavpts (Refpts):        Cavity location for off-momentum tuning

    Returns:
        elemdata0:      Linear optics data at the entrance of the ring
        ringdata:       Lattice properties
        elemdata:       Linear optics at the points refered to by *refpts*,
          if refpts is :py:obj:`None` an empty lindata structure is returned.

    **elemdata** is a record array with fields:

    ================    ===================================================
    **s_pos**           longitudinal position [m]
    **M**               (6, 6) transfer matrix M from the beginning of ring
                        to the entrance of the element
    **closed_orbit**    (6,) closed orbit vector
    **dispersion**      (4,) dispersion vector
    **A**               (6,6) A-matrix
    **R**               (3, 6, 6) R-matrices
    **beta**            :math:`\left[ \beta_x,\beta_y \right]` vector
    **alpha**           :math:`\left[ \alpha_x,\alpha_y \right]` vector
    **mu**              :math:`\left[ \mu_x,\mu_y \right]`, betatron phase
                        (modulo :math:`2\pi`)
    **W**               :math:`\left[ W_x,W_y \right]` only if *get_w*
                        is :py:obj:`True`: chromatic amplitude function
    **Wp**              :math:`\left[ Wp_x,Wp_y \right]` only if *get_w*
                        is :py:obj:`True`: chromatic phase function
    **dalpha**          (2,) alpha derivative vector
                        (:math:`\Delta \alpha/ \delta_p`)
    **dbeta**           (2,) beta derivative vector
                        (:math:`\Delta \beta/ \delta_p`)
    **dmu**             (2,) mu derivative vector
                        (:math:`\Delta \mu/ \delta_p`)
    **ddispersion**     (4,) dispersion derivative vector
                        (:math:`\Delta D/ \delta_p`)
    **dR**              (3, 6, 6) R derivative vector
                        (:math:`\Delta R/ \delta_p`)
    ================    ===================================================

    All values given at the entrance of each element specified in refpts.
    Field values can be obtained with either
    *lindata['idx']* or *lindata.idx*

    **ringdata** is a record array with fields:

    =================   ======
    **tune**            Fractional tunes
    **chromaticity**    Chromaticities, only computed if *get_chrom* is
                        :py:obj:`True`
    **damping_time**    Damping times [s] (only if radiation is ON)
    =================   ======

    References:
        **[9]** Etienne Forest, Phys. Rev. E 58, 2481 – Published 1 August 1998

        **[10]** Andrzej Wolski, Phys. Rev. ST Accel. Beams 9, 024001 –
        Published 3 February 2006

        .. [11] Brian W. Montague Report LEP Note 165, CERN, 1979
    """
    return _linopt(ring, _analyze6, *args, **kwargs)


def linopt_auto(ring: Lattice, *args, **kwargs):
    """
    This is a convenience function to automatically switch to the faster
    :py:func:`linopt2` in case the *coupled* keyword argument is
    :py:obj:`False` **and** ring.is_6d is :py:obj:`False`.
    Otherwise the default :py:func:`linopt6` is used

    Parameters: Same as :py:func:`.linopt2` or :py:func:`.linopt6`

    Keyword Args;
        coupled (bool):     If set to :py:obj:`False`, H/V coupling
          will be ignored to simplify the calculation (needs ring.is_6d
          :py:obj:`False`)


    Returns:
        elemdata0:      Linear optics data at the entrance of the ring
        ringdata:       Lattice properties
        elemdata:       Linear optics at the points refered to by *refpts*,
          if refpts is :py:obj:`None` an empty lindata structure is returned.

    Warning:
        The output varies depending whether :py:func:`.linopt2` or
        :py:func:`.linopt6` is called. To be used with care!
    """
    if not (kwargs.pop('coupled', True) or ring.is_6d):
        return linopt2(ring, *args, **kwargs)
    else:
        return linopt6(ring, *args, **kwargs)


def get_optics(ring: Lattice, refpts: Refpts = None,
               dp: float = None,
               method: Callable = linopt6,
               **kwargs):
    """Linear analysis of a fully coupled lattice

    Parameters:
        ring:                   Lattice description.
        refpts:                 Elements at which data is returned.
          It can be:

          1. an integer in the range [-len(ring), len(ring)-1]
             selecting the element according to python indexing rules.
             As a special case, len(ring) is allowed and refers to the end
             of the last element,
          2. an ordered list of such integers without duplicates,
          3. a numpy array of booleans of maximum length len(ring)+1,
             where selected elements are :py:obj:`True`.
        dp:                     Momentum deviation.
        method (Callable):      Method for linear optics:

          :py:obj:`~.linear.linopt2`: no longitudinal motion, no H/V coupling,

          :py:obj:`~.linear.linopt4`: no longitudinal motion, Sagan/Rubin
          4D-analysis of coupled motion,

          :py:obj:`~.linear.linopt6` (default): with or without longitudinal
          motion, normal mode analysis

    Keyword Args:
        dct (float):            Path lengthening. Defaults to :py:obj:`None`
        df (float):             Deviation of RF frequency. Defaults to
          :py:obj:`None`
        orbit (Orbit):          Avoids looking for the closed orbit if it is
          already known ((6,) array)
        get_chrom (bool):       Compute chromaticities. Needs computing
          the tune at 2 different momentum deviations around the central one.
        get_w (bool):           Computes chromatic amplitude functions
          (W) [4]_. Needs to compute the optics at 2 different momentum
          deviations around the central one.
        keep_lattice (bool):    Assume no lattice change since the
          previous tracking. Defaults to :py:obj:`False`
        XYStep (float):         Step size.
          Default: :py:data:`DConstant.XYStep <.DConstant>`
        DPStep (float):         Momentum step size.
          Default: :py:data:`DConstant.DPStep <.DConstant>`
        twiss_in:               Initial conditions for transfer line optics.
          Record array as output by :py:func:`.linopt6`, or dictionary. Keys:

          R or alpha, beta
            mandatory (2,) arrays
          closed_orbit
            Optional (6,) array, default 0
          dispersion
            Optional (6,) array, default 0

          If present, the attribute **R** will be used, otherwise the
          attributes **alpha** and **beta** will be used. All other attributes
          are ignored.

    Returns:
        elemdata0:      Linear optics data at the entrance of the ring
        ringdata:       Lattice properties
        elemdata:       Linear optics at the points refered to by *refpts*,
          if refpts is :py:obj:`None` an empty lindata structure is returned.

    Warning:
        The format of output record arrays depends on the selected method.
        See :py:func:`linopt2`, :py:func:`linopt4`, :py:func:`linopt6`.
    """
    return method(ring, refpts=refpts, dp=dp, **kwargs)


# noinspection PyPep8Naming
@check_6d(False)
def linopt(ring: Lattice, dp: float = 0.0, refpts: Refpts = None,
           get_chrom: bool = False, **kwargs):
    """Linear analysis of a H/V coupled lattice (deprecated)

    Parameters:
        ring:           lattice description.
        dp:             momentum deviation.
        refpts:         elements at which data is returned. It can be:
                        1) an integer in the range [-len(ring), len(ring)-1]
                           selecting the element according to python indexing
                           rules. As a special case, len(ring) is allowed and
                           refers to the end of the last element,
                        2) an ordered list of such integers without duplicates,
                        3) a numpy array of booleans of maximum length
                           len(ring)+1, where selected elements are
                           :py:obj:`True`.
    Keyword Args:
        orbit           avoids looking for the closed orbit if is already known
                        ((6,) array)
        get_chrom=False compute chromaticities. Needs computing the tune at
                        2 different momentum deviations around the central one.
        get_w=False     computes chromatic amplitude functions (W) [4].
                        Needs to compute the optics at 2 different momentum
                        deviations around the central one.
        keep_lattice    Assume no lattice change since the previous tracking.
                        Defaults to :py:obj:`False`
        XYStep=1.0e-8   transverse step for numerical computation
        DPStep=1.0E-6   momentum deviation used for computation of
                        chromaticities and dispersion
        coupled=True    if :py:obj:`False`, simplify the calculations by
                        assuming no H/V coupling
        twiss_in=None   Initial conditions for transfer line optics. Record
                        array as output by linopt, or dictionary. Keys:
                        'alpha' and 'beta'  (mandatory)
                        'closed_orbit',     (default 0)
                        'dispersion'        (default 0)
                        All other attributes are ignored.
    OUTPUT
        lindata0        linear optics data at the entrance of the ring
        tune            [tune_A, tune_B], linear tunes for the two normal modes
                        of linear motion [1]
        chrom           [ksi_A , ksi_B], chromaticities ksi = d(nu)/(dP/P).
                        Only computed if 'get_chrom' is :py:obj:`True`
        lindata         linear optics at the points refered to by refpts, if
                        refpts is None an empty lindata structure is returned.

        lindata is a record array with fields:
        idx             element index in the ring
        s_pos           longitudinal position [m]
        m44             (4, 4) transfer matrix M from the beginning of ring
                        to the entrance of the element [2]
        closed_orbit    (6,) closed orbit vector
        dispersion      (4,) dispersion vector
        beta            [betax, betay] vector
        alpha           [alphax, alphay] vector
        mu              [mux, muy], betatron phase (modulo 2*pi)
        W               (2,) chromatic amplitude function (only if get_w==True)
        All values given at the entrance of each element specified in refpts.
        In case coupled == :py:obj:`True` additional outputs are available:
        gamma           gamma parameter of the transformation to eigenmodes
        A               (2, 2) matrix A in [3]
        B               (2, 2) matrix B in [3]
        C               (2, 2) matrix C in [3]
        Field values can be obtained with either
        lindata['idx']    or
        lindata.idx
    REFERENCES
        [1] D.Edwards,L.Teng IEEE Trans.Nucl.Sci. NS-20, No.3, p.885-888, 1973
        [2] E.Courant, H.Snyder
        [3] D.Sagan, D.Rubin Phys.Rev.Spec.Top.-Accelerators and beams,
            vol.2 (1999)
        [4] Brian W. Montague Report LEP Note 165, CERN, 1979

    :meta private:
    """
    analyze = _analyze4 if kwargs.pop('coupled', True) else _analyze2
    eld0, bd, eld = _linopt(ring, analyze, refpts, dp=dp, get_chrom=get_chrom,
                            add0=(0,), adds=(get_uint32_index(ring, refpts),),
                            addtype=[('idx', numpy.uint32)],
                            mname='m44', **kwargs)
    return eld0, bd.tune, bd.chromaticity, eld


# noinspection PyPep8Naming
@check_6d(False)
def avlinopt(ring, dp, refpts, **kwargs):
    r"""Linear analysis of a lattice with average values

    :py:func:`avlinopt` returns average beta, mu, dispersion over the lattice
    elements.

    Parameters:
        ring:       Lattice description.
        dp:         Momentum deviation.
        refpts:     Elements at which data is returned.
          It can be:

          1. an integer in the range [-len(ring), len(ring)-1]
             selecting the element according to python indexing rules.
             As a special case, len(ring) is allowed and refers to the end
             of the last element,
          2. an ordered list of such integers without duplicates,
          3. a numpy array of booleans of maximum length len(ring)+1,
             where selected elements are :py:obj:`True`.

    Keyword Args:
        dct (float):            Path lengthening. Defaults to :py:obj:`None`
        df (float):             Deviation of RF frequency. Defaults to
          :py:obj:`None`
        orbit (Orbit):          Avoids looking for the closed orbit if it is
          already known ((6,) array)
        get_chrom (bool):       Compute chromaticities. Needs computing
          the tune at 2 different momentum deviations around the central one.
        get_w (bool):           Computes chromatic amplitude functions
          (W) [4]_. Needs to compute the optics at 2 different momentum
          deviations around the central one.
        keep_lattice (bool):    Assume no lattice change since the
          previous tracking. Defaults to :py:obj:`False`
        XYStep (float):         Step size.
          Default: :py:data:`DConstant.XYStep <.DConstant>`
        DPStep (float):         Momentum step size.
          Default: :py:data:`DConstant.DPStep <.DConstant>`
        twiss_in:               Initial conditions for transfer line optics.
          Record array as output by :py:func:`.linopt6`, or dictionary. Keys:

          R or alpha, beta
            mandatory (2,) arrays
          closed_orbit
            Optional (6,) array, default 0
          dispersion
            Optional (6,) array, default 0

          If present, the attribute **R** will be used, otherwise the
          attributes **alpha** and **beta** will be used. All other attributes
          are ignored.

    Returns:
        elemdata:   Linear optics at the points refered to by *refpts*,
          if refpts is :py:obj:`None` an empty lindata structure is returned.
        avebeta:    Average beta functions
          [:math:`\hat{\beta_x},\hat{\beta_y}`] at *refpts*
        avemu:      Average phase advances
          [:math:`\hat{\mu_x},\hat{\mu_y}`] at *refpts*
        avedisp:    Average dispersion [:math:`\hat{\eta_x}, \hat{\eta'_x},
          \hat{\eta_y}, \hat{\eta'_y}`] at *refpts*
        avespos:    Average s position at *refpts*
        tune:       [:math:`\nu_1,\nu_2`], linear tunes for the two normal
          modes of linear motion [1]
        chrom:      [:math:`\xi_1,\xi_2`], chromaticities

    See also:
        :py:func:`linopt4`, :py:func:`get_optics`
    """
    def get_strength(elem):
        try:
            k = elem.PolynomB[1]
        except (AttributeError, IndexError):
            k = numpy.finfo(numpy.float64).eps
        return k

    def get_sext_strength(elem):
        try:
            k = elem.PolynomB[2]
        except (AttributeError, IndexError):
            k = 0.0
        return k

    def get_roll(elem):
        try:
            k = elem.R2[0][0]
        except (AttributeError, IndexError):
            k = 1.0
        return k

    def get_dx(elem):
        try:
            k = (elem.T2[0]-elem.T1[0])/2.0
        except (AttributeError, IndexError):
            k = 0.0
        return k

    def get_bendingangle(elem):
        try:
            k = elem.BendingAngle
        except (AttributeError, IndexError):
            k = 0.0
        return k

    def get_e1(elem):
        try:
            k = elem.EntranceAngle
        except (AttributeError, IndexError):
            k = 0.0
        return k

    def get_fint(elem):
        try:
            k = elem.FringeInt1
        except (AttributeError, IndexError):
            k = 0.0
        return k

    def get_gap(elem):
        try:
            k = elem.FullGap
        except (AttributeError, IndexError):
            k = 0.0
        return k

    def sini(x, L):
        r = x.copy()
        r[x > 0] = numpy.sin(numpy.sqrt(x[x > 0])*L[x > 0]
                             )/numpy.sqrt(x[x > 0])
        r[x < 0] = numpy.sinh(numpy.sqrt(-x[x < 0])*L[x < 0]
                              )/numpy.sqrt(-x[x < 0])
        return r

    def cosi(x, L):
        r = x.copy()
        r[x > 0] = numpy.cos(numpy.sqrt(x[x > 0])*L[x > 0])
        r[x < 0] = numpy.cosh(numpy.sqrt(-x[x < 0])*L[x < 0])
        return r

    def betadrift(beta0, alpha0, lg):
        gamma0 = (alpha0 * alpha0 + 1.0) / beta0
        return beta0-alpha0*lg+gamma0*lg*lg/3

    def betafoc(beta0, alpha0, kk, lg):
        gamma0 = (alpha0 * alpha0 + 1.0) / beta0
        return ((beta0+gamma0/kk)*lg +
                (beta0-gamma0/kk)*sini(kk, 2.0*lg)/2.0 +
                (cosi(kk, 2.0*lg)-1.0)*alpha0/kk)/lg/2.0

    def dispfoc(disp0, ir, k2, lg):
        avedisp = disp0.copy()
        avedisp[:, 0::2] = (disp0[:, 0::2]*(sini(k2, lg)) +
                            disp0[:, 1::2]*(1.0-cosi(k2, lg))/k2 +
                            ir*(lg-sini(k2, lg))/k2)/lg
        avedisp[:, 1::2] = (disp0[:, 0::2]*(sini(k2, lg)) -
                            disp0[:, 0::2]*(1.0-cosi(k2, lg)) +
                            ir*(lg-cosi(k2, lg))/k2)/lg
        return avedisp

    # selected list
    boolrefs = get_bool_index(ring, refpts)
    ring_selected = ring[refpts]
    L = numpy.array([el.Length for el in ring_selected])
    K = numpy.array([get_strength(el) for el in ring_selected])
    sext_strength = numpy.array([get_sext_strength(el)
                                 for el in ring_selected])
    roll = numpy.array([get_roll(el) for el in ring_selected])
    ba = numpy.array([get_bendingangle(el) for el in ring_selected])
    e1 = numpy.array([get_e1(el) for el in ring_selected])
    Fint = numpy.array([get_fint(el) for el in ring_selected])
    gap = numpy.array([get_gap(el) for el in ring_selected])
    dx = numpy.array([get_dx(el) for el in ring_selected])
    irho = ba.copy()
    d_csi = ba.copy()

    # whole ring list
    longelem = get_bool_index(ring, None)
    longelem[boolrefs] = (L != 0)

    shorti_refpts = (~longelem) & boolrefs
    longi_refpts = longelem & boolrefs
    longf_refpts = numpy.roll(longi_refpts, 1)
    all_refs = shorti_refpts | longi_refpts | longf_refpts
    _, bd, d_all = linopt4(ring, refpts=all_refs, dp=dp,
                           get_chrom=True, **kwargs)
    lindata = d_all[boolrefs[all_refs]]

    avebeta = lindata.beta.copy()
    avemu = lindata.mu.copy()
    avedisp = lindata.dispersion.copy()
    aves = lindata.s_pos.copy()
    ClosedOrbit = lindata.closed_orbit.copy()

    di = d_all[longi_refpts[all_refs]]
    df = d_all[longf_refpts[all_refs]]

    b_long = (L != 0.0)
    b_foc = (K != 0.0)
    b_foc_long = b_long & b_foc
    b_drf = b_long & (~b_foc)

    dx0 = ClosedOrbit[:, 0]
    dx0[b_long] = (di.closed_orbit[:, 0]+df.closed_orbit[:, 0])/2
    K = K*roll + 2*sext_strength*(dx0-dx)

    K2 = numpy.stack((K[b_foc_long], -K[b_foc_long]), axis=1)
    fff = b_foc_long[b_long]
    irho[b_long] = ba[b_long]/L[b_long]
    d_csi[b_long] = ba[b_long]*(gap[b_long]*Fint[b_long] *
                                (1.0 + numpy.sin(e1[b_long])**2) /
                                numpy.cos(e1[b_long])/L[b_long])
    L2 = numpy.stack((L[b_foc_long], L[b_foc_long]), axis=1)
    irho = irho.reshape((-1, 1))
    L = L.reshape((-1, 1))

    avemu[b_long] = 0.5 * (di.mu + df.mu)
    aves[b_long] = 0.5 * (df.s_pos + di.s_pos)
    avebeta[b_drf] = betadrift(di.beta[~fff], di.alpha[~fff], L[b_drf])
    avebeta[b_foc_long] = betafoc(di.beta[fff], di.alpha[fff], K2, L2)
    avedisp[numpy.ix_(b_long, [1, 3])] = (df.dispersion[:, [0, 2]] -
                                          di.dispersion[:, [0, 2]]) / L[b_long]
    idx = numpy.ix_(~fff, [0, 2])
    avedisp[numpy.ix_(b_drf, [0, 2])] = (di.dispersion[idx] +
                                         df.dispersion[idx]) * 0.5
    avedisp[b_foc_long, :] = dispfoc(di.dispersion[fff, :],
                                     irho[b_foc_long], K2, L2)

    return lindata, avebeta, avemu, avedisp, aves, bd.tune, bd.chromaticity


@frequency_control
def get_tune(ring: Lattice, *, method: str = 'linopt',
             dp: float = None, dct: float = None, df: float = None,
             orbit: Orbit = None, **kwargs):
    r"""Computes the tunes using several available methods

    Parameters:
        ring:                   Lattice description
        method:                 ``'linopt'`` returns the tunes from the
          :py:func:`.linopt6` function,

          ``'fft'`` tracks a single particle and computes the tunes with fft,

          ``'interp_fft'`` tracks a single particle and computes the tunes with
          interpolated FFT.
        dp (float):             Momentum deviation.
        dct (float):            Path lengthening.
        df (float):             Deviation of RF frequency.
        orbit (Orbit):          Avoids looking for the closed orbit if it is
          already known ((6,) array)

    for the ``'fft'`` and ``'interp_fft'`` methods only:

    Keyword Args:
        nturns (int):           Number of turns. Default: 512
        amplitude (float):      Amplitude of oscillation.
          Default: 1.E-6
        remove_dc (bool):       Remove the mean of oscillation data.
          Default: :py:obj:`True`
        num_harmonics (int):    Number of harmonic components to
          compute (before mask applied, default: 20)
        fmin (float):           Lower tune bound. Default: 0
        fmax (float):           Upper tune bound. Default: 1
        hann (bool):            Turn on Hanning window.
          Default: :py:obj:`False`. Work only for ``'fft'``
        get_integer(bool):   Turn on integer tune (slower)

    Returns:
        tunes (ndarray):        array([:math:`\nu_x,\nu_y`])
    """
    # noinspection PyShadowingNames
    def gen_centroid(ring, ampl, nturns, remove_dc, ld):
        nv = ld.A.shape[0]
        p0 = ld.closed_orbit.copy()
        p0[0] += ampl
        p0[2] += ampl
        if nv >= 6:
            p0[4] += ampl
        p1 = numpy.squeeze(internal_lpass(ring, p0, nturns, len(ring)))
        if remove_dc:
            p1 -= numpy.mean(p1, axis=1, keepdims=True)
        p2 = solve(ld.A, p1[:nv, :])
        return numpy.conjugate(p2.T.view(dtype=complex).T)
    get_integer = kwargs.pop('get_integer', False)
    if get_integer:
        assert method == 'linopt', \
           'Integer tune only accessible with method=linopt'
    if method == 'linopt':
        if get_integer:
            _, _, c = get_optics(ring, refpts=range(len(ring)+1),
                                 dp=dp, dct=dct, df=df, orbit=orbit)
            tunes = c.mu[-1]/(2*numpy.pi)
        else:
            tunes = _tunes(ring, dp=dp, dct=dct, df=df, orbit=orbit)
    else:
        nturns = kwargs.pop('nturns', 512)
        ampl = kwargs.pop('ampl', 1.0e-6)
        remove_dc = kwargs.pop('remove_dc', True)
        ld, _, _ = linopt6(ring, dp=dp, dct=dct, df=df, orbit=orbit)
        cents = gen_centroid(ring, ampl, nturns, remove_dc, ld)
        tunes = get_tunes_harmonic(cents, method=method, **kwargs)
    return tunes


@frequency_control
def get_chrom(ring: Lattice, *, method: str = 'linopt',
              dp: float = None, dct: float = None, df: float = None,
              cavpts: Refpts = None, **kwargs):
    r"""Computes the chromaticities using several available methods

    Parameters:
        ring:               Lattice description.
        method:             ``'linopt'`` returns the tunes from the
          :py:func:`linopt6` function,

          ``'fft'`` tracks a single particle and computes the tunes with
          :py:func:`~scipy.fftpack.fft`,

          ``'interp_fft'`` tracks a single particle and computes the tunes with
          interpolated FFT.
        dp (float):         Momentum deviation.
        dct (float):        Path lengthening.
        df (float):         Deviation of RF frequency.
        cavpts:     If :py:obj:`None`, look for ring.cavpts, or
          otherwise take all cavities.

    Keyword Args:
        DPStep (float):     Momentum step for differentiation
          Default: :py:data:`DConstant.DPStep <.DConstant>`

    for the ``'fft'`` and ``'interp_fft'`` methods only:

    Keyword Args:
        nturns (int):       Number of turns. Default: 512
        amplitude (float):  Amplitude of oscillation.
          Default: 1.E-6
        remove_dc (bool):   Remove the mean of oscillation data.
          Default: :py:obj:`True`
        num_harmonics (int):Number of harmonic components to
          compute (before mask applied, default: 20)
        fmin (float):       Lower tune bound. Default: 0
        fmax (float):       Upper tune bound. Default: 1
        hann (bool):        Turn on Hanning window.
          Default: :py:obj:`False`, Work only for ``'fft'``

    Returns:
        chromaticities (ndarray):   array([:math:`\xi_x,\xi_y`])
    """

    dp_step = kwargs.pop('DPStep', DConstant.DPStep)
    if method == 'fft':
        print('Warning fft method not accurate to get the ' +
              'chromaticity')

    if ring.is_6d:
        f0 = ring.get_rf_frequency(cavpts=cavpts)
        df = dp_step * ring.disable_6d(copy=True).slip_factor * f0
        rgup = ring.set_rf_frequency(f0 + 0.5 * df, cavpts=cavpts, copy=True)
        o0up, _ = find_orbit6(rgup, **kwargs)
        tune_up = get_tune(rgup,  method=method, orbit=o0up, **kwargs)
        rgdn = ring.set_rf_frequency(f0 - 0.5 * df, cavpts=cavpts, copy=True)
        o0dn, _ = find_orbit6(rgdn, **kwargs)
        tune_down = get_tune(rgdn,  method=method, orbit=o0dn, **kwargs)
        dp_step = o0up[4] - o0dn[4]
    else:
        if dct is not None or df is not None:
            dp = find_orbit4(ring, dct=dct, df=df)[0][4]
        elif dp is None:
            dp = 0.0
        tune_up = get_tune(ring, method=method, dp=dp + 0.5*dp_step, **kwargs)
        tune_down = get_tune(ring, method=method,
                             dp=dp - 0.5*dp_step, **kwargs)

    return (tune_up - tune_down) / dp_step


Lattice.linopt = linopt
Lattice.linopt2 = linopt2
Lattice.linopt4 = linopt4
Lattice.linopt6 = linopt6
Lattice.get_optics = get_optics
Lattice.avlinopt = avlinopt
Lattice.get_tune = get_tune
Lattice.get_chrom = get_chrom



import numpy as np
import at
from at.constants import clight
import matplotlib.pyplot as plt
import pickle as pkl
from scipy.optimize import curve_fit


def get_growth_rates():
    '''
    Function to compute mode 0 TRW mode growth rates
    R. Nagaoka, K. Bane, Collective Effects in Diffraction
    limited storage rings, eq31

    '''

    omega0 = 2*np.pi*ring.revolution_frequency
    energy = ring.energy
    R = ring.circumference/(2*np.pi)

    gr = (betax*omega0*current*R) / (4*np.pi*energy*beff**3) \
        * ((2*clight*Z0*rho_material) / ((1-qx)*omega0))**0.5

    return gr


def sin_fit(x, amp, freq, phase, off):
    return amp * np.sin(freq * x + phase) + off


def getTimeVectors(tbtMatrix):

    print('Do SVD')
    u, s, v = np.linalg.svd(tbtMatrix)

    # time vector of strongest mode
    vector = np.array([v[0][k] for k in np.arange(v.shape[1])])

    # SVD does not give mode number, so we fit a sin to the spatial pattern
    fit, _ = curve_fit(sin_fit, np.arange(Nbunches), np.real(u[:, 0]),
                       p0=[np.amax(np.real(u[:, 0])),
                       -2*np.pi/Nbunches,
                       -np.pi/4, 0])
    mode = fit[1]/2/np.pi*Nbunches

    bunchPlot = np.arange(0, Nbunches, 0.01)

    fig = plt.figure()
    ax1 = fig.add_subplot(111)
    ax1.plot(np.real(u[:, 0]),
             color='r', marker='x', linestyle='None', label='Spatial Pattern')
    ax1.plot(bunchPlot, sin_fit(bunchPlot, fit[0], fit[1], fit[2], fit[3]),
             color='b', linestyle='solid', label='Sinusoidal Fit')
    ax1.set_xlabel('Bunch')
    ax1.legend()
    plt.show()

    return vector, mode, fit


def fitTimeVectors(tv):

    width = 400
    step = 200

    numsteps = int((len(tv) - width)/step) + 1
    starts = np.arange(0, len(tv) - width + 1, step)
    ends = np.arange(width, len(tv) + 1, step)

    dat = np.real(np.abs(tv))

    newdat = np.array([np.amax(dat[starts[i]:ends[i]])
                       for i in np.arange(numsteps)])

    turns = np.arange(len(dat))
    fitVal = np.polyfit((starts+ends) / 2, np.log(newdat), 1)
    chi_squared = np.sum((
                  np.polyval(fitVal, (starts+ends)/2) - np.log(newdat))**2)

    if 1/fitVal[0] > 1e6:
        fitVal[0] = np.inf

    return fitVal, newdat, starts, ends, chi_squared


# First we define everything we need and get the theoretical growth rates
Z0 = 377.7  # Vacuum impedance [Ohm]

resDict = pkl.load(open('./TRW_output.pkl', 'rb'))
ring = resDict['ring']
Nbunches = resDict['M']
current = resDict['current']
beff = resDict['rvac']
rho_material = 1/resDict['sigmac']


ld0, bd, ld = ring.get_optics()
betax = ld0.beta[0]
alphax = ld0.alpha[0]

f0 = ring.revolution_frequency
alpha = at.get_mcf(ring.radiation_off(copy=True))
eta = at.get_slip_factor(ring.radiation_off(copy=True))
envel = at.envelope_parameters(ring.radiation_on(copy=True))
qx, qy, qs = ring.get_tune()
fs0 = qs*f0

gr_theory = get_growth_rates()

# Find the last turn in the data where there were no lost particles
tstart = 4000
try:
    tend = np.where(not resDict['lostturn'])[0][0]
except:
    tend = len(resDict['xp'])
tend = 7000

data_xp = resDict['xp'].T
data_x = resDict['x'].T

# Plot all data dp, just to see how it looks
bbs = np.arange(0, Nbunches)
for bb in bbs:
    plt.plot(data_x[bb, tstart:tend])
plt.show()

# Here we use xp and x and combine with betax into one complex number
# this allows us to distinguish from + and - modes
inv = np.array([[1/np.sqrt(betax), 0],
                [alphax/np.sqrt(betax), np.sqrt(betax)]])

data = np.zeros(data_xp[:, tstart:tend].shape) + 1j * 0
for i in np.arange(data_xp.shape[0]):
    xp = data_xp[i, tstart:tend]
    x = data_x[i, tstart:tend]
    datpos, datmom = np.matmul(inv, [x, xp])
    data[i, :] = datpos + 1j * datmom

# Do the SVD and get all the time vectors
tv, mode, _ = getTimeVectors(data)
mode = int(np.round(mode))

# Fit the time vector
gr, newdat, starts, ends, chis = fitTimeVectors(tv)

turns = np.arange(0, tv.shape[0], 50)

fig = plt.figure()
ax1 = fig.add_subplot(111)
ax1.plot(np.real(tv), color='r', linestyle='solid', marker='None')
ax1.plot(turns, np.exp(gr[1])*np.exp(turns*gr[0]),
         linestyle='solid', marker='None',
         label=r'$\tau_{{\mathrm{{sim}}}}={:.2f}\ \mathrm{{turns}}$'
               .format(1/gr[0]) + '\n' +
               r'$\tau_{{\mathrm{{analytic}}}}={:.2f}\ \mathrm{{turns}}$'
               .format(f0/gr_theory))
ax1.plot((starts + ends)/2, newdat, color='k', marker='x', linestyle='None')
ax1.legend()
ax1.set_title('Mode {:d}, mu={:.3f}'.format(mode, mode/Nbunches))
ax1.set_xlabel('Turn')
plt.show()



"""Variables are **references** to scalar attributes of lattice elements. There are 2
kinds of element variables:

- an :py:class:`ElementVariable` is associated to an element object, and acts on all
  occurences of this object. But it will not affect any copy, neither shallow nor deep,
  of the original object,
- a :py:class:`RefptsVariable` is not associated to an element object, but to an element
  location in a :py:class:`.Lattice`. It acts on any copy of the initial lattice. A
  *ring* argument must be provided to the *set* and *get* methods to identify the
  lattice, which may be a possibly modified copy of the original lattice
"""

from __future__ import annotations

__all__ = ["RefptsVariable", "ElementVariable"]

from collections.abc import Sequence
from typing import Union, Optional

import numpy as np

from .elements import Element
from .lattice_object import Lattice
from .utils import Refpts, getval, setval
from .variables import VariableBase


class RefptsVariable(VariableBase):
    r"""A reference to a scalar attribute of :py:class:`.Lattice` elements.

    It can refer to:

    * a scalar attribute or
    * an element of an array attribute

    of one or several :py:class:`.Element`\ s of a lattice.

    A :py:class:`RefptsVariable` is not associated to element objets themselves, but
    to the location of these elements in a lattice. So a :py:class:`RefptsVariable`
    will act equally on any copy of the initial ring.
    As a consequence, a *ring* keyword argument (:py:class:`.Lattice` object) must be
    supplied for getting or setting the variable.
    """

    def __init__(
        self, refpts: Refpts, attrname: str, index: Optional[int] = None, **kwargs
    ):
        r"""
        Parameters:
            refpts:     Location of variable :py:class:`.Element`\ s
            attrname:   Attribute name
            index:      Index in the attribute array. Use :py:obj:`None` for
              scalar attributes

        Keyword Args:
            name (str):     Name of the Variable. Default: ``''``
            bounds (tuple[float, float]):   Lower and upper bounds of the
              variable value. Default: (-inf, inf)
            delta (float):  Step. Default: 1.0
            ring (Lattice): If specified, it is used to get and store the initial
              value of the variable. Otherwise, the initial value is set to None
        """
        self._getf = getval(attrname, index=index)
        self._setf = setval(attrname, index=index)
        self.refpts = refpts
        super().__init__(**kwargs)

    def _setfun(self, value: float, ring: Lattice = None):
        if ring is None:
            raise ValueError("Can't set values if ring is None")
        for elem in ring.select(self.refpts):
            self._setf(elem, value)

    def _getfun(self, ring: Lattice = None) -> float:
        if ring is None:
            raise ValueError("Can't get values if ring is None")
        values = np.array([self._getf(elem) for elem in ring.select(self.refpts)])
        return np.average(values)


class ElementVariable(VariableBase):
    r"""A reference to a scalar attribute of :py:class:`.Lattice` elements.

    It can refer to:

    * a scalar attribute or
    * an element of an array attribute

    of one or several :py:class:`.Element`\ s of a lattice.

    An :py:class:`ElementVariable` is associated to an element object, and acts on all
    occurrences of this object. But it will not affect any copy, neither shallow nor
    deep, of the original object.
    """

    def __init__(
        self,
        elements: Union[Element, Sequence[Element]],
        attrname: str,
        index: Optional[int] = None,
        **kwargs,
    ):
        r"""
        Parameters:
            elements:   :py:class:`.Element` or Sequence[Element] whose
              attribute is varied
            attrname:   Attribute name
            index:      Index in the attribute array. Use :py:obj:`None` for
              scalar attributes

        Keyword Args:
            name (str):     Name of the Variable. Default: ``''``
            bounds (tuple[float, float]):   Lower and upper bounds of the
              variable value. Default: (-inf, inf)
            delta (float):  Step. Default: 1.0
        """
        # Ensure the uniqueness of elements
        if isinstance(elements, Element):
            self._elements = {elements}
        else:
            self._elements = set(elements)
        self._getf = getval(attrname, index=index)
        self._setf = setval(attrname, index=index)
        super().__init__(**kwargs)

    def _setfun(self, value: float, **kwargs):
        for elem in self._elements:
            self._setf(elem, value)

    def _getfun(self, **kwargs) -> float:
        values = np.array([self._getf(elem) for elem in self._elements])
        return np.average(values)

    @property
    def elements(self):
        """Return the set of elements acted upon by the variable"""
        return self._elements



"""Implements PEP 517 and PEP 660 hooks for custom build"""

import os

from setuptools import build_meta as _orig
from setuptools.build_meta import *  # noqa: F401, F403


def _get_option(config, key, default_value):
    val = os.environ.get(key.upper(), default_value)
    if config is not None:
        val = config.get(key.lower(), val)
    return eval(val)


def _set_environ(func):
    """Set environment for building according to config"""

    def build(wheel_dir, config_settings=None, metadata_dir=None):
        print("** Entering set_environ. " "Config settings:", config_settings)
        os.environ["MPI"] = str(_get_option(config_settings, "MPI", "0"))
        os.environ["OPENMP"] = str(_get_option(config_settings, "OPENMP", "0"))
        omp_threshold = _get_option(config_settings, "OMP_PARTICLE_THRESHOLD", "None")
        if omp_threshold is not None:
            os.environ["OMP_PARTICLE_THRESHOLD"] = str(omp_threshold)
        print("** MPI:", os.environ.get("MPI", "None"))
        print("** OPENMP:", os.environ.get("OPENMP", "None"))
        ret = func(wheel_dir, config_settings, metadata_dir)
        print("** Leaving set_environ")
        return ret

    return build


def _requirements(func):
    """Add build requirements according to config"""

    def get_requires(config_settings=None):
        print("** Entering get_requires. Config settings:", config_settings)
        mpi = _get_option(config_settings, "MPI", "0")
        print("** MPI:", mpi)
        addlist = ["mpi4py"] if mpi else []
        print("** Additional modules:", addlist)
        ret = func(config_settings) + addlist
        print("** Leaving get_requires")
        return ret

    return get_requires


build_wheel = _set_environ(_orig.build_wheel)
build_editable = _set_environ(_orig.build_editable)

get_requires_for_build_wheel = _requirements(_orig.get_requires_for_build_wheel)
get_requires_for_build_editable = _requirements(_orig.get_requires_for_build_editable)



from .elements import Element
import numpy
import io
from ..constants import clight
from warnings import warn


def _anyarray(value):
    # Ensure proper ordering(F) and alignment(A) for "C" access in integrators
    return numpy.require(value, dtype=numpy.float64, requirements=['F', 'A'])


class InsertionDeviceKickMap(Element):
    """
    Insertion Device Element. Valid for a parallel electron beam.

      Pascale ELLEAUME, "A New Approach to the Electron Beam Dynamics in
        Undulators and  Wigglers". EPAC1992 0661.
        European Synchrotron Radiation Facility.
        BP 220, F-38043 Grenoble, France
    """
    _BUILD_ATTRIBUTES = Element._BUILD_ATTRIBUTES + ['PassMethod',
                                                     'Filename_in',
                                                     'Normalization_energy',
                                                     'Nslice',
                                                     'Length',
                                                     'xkick',
                                                     'ykick',
                                                     'xkick1',
                                                     'ykick1',
                                                     'xtable',
                                                     'ytable']

    _conversions = dict(Element._conversions,
                        Nslice=int,
                        xkick=_anyarray,
                        ykick=_anyarray,
                        xkick1=_anyarray,
                        ykick1=_anyarray
                        )

    def set_DriftPass(self):
        setattr(self, 'PassMethod', 'DriftPass')

    def set_IdTablePass(self):
        setattr(self, 'PassMethod', 'IdTablePass')

    def get_PassMethod(self):
        warn(UserWarning('get_PassMethod is deprecated; do not use'))
        return getattr(self, 'PassMethod')

    def from_text_file(self, Nslice, Filename_in, Energy):
        """
        This function creates an Insertion Device Kick Map
        from a Radia field map file.

        Args:
            family_name:    family name
            Nslice:         number of slices in integrator
            Filename_in:    input filename
            Energy:         particle energy in GeV

        Returns:
            KickMap element
            Default PassMethod: ``IdTablePass``
        """
        # 2023jul04 changing class to save in .mat and .m
        # 2023apr30 redefinition to function
        # 2023jan18 fix bug with element print
        # 2023jan15 first release
        # orblancog
        def readRadiaFieldMap(file_in_name):
            """
            Read a RadiaField map and return
            """
            with io.open(file_in_name, mode="r", encoding="utf-8") as f:
                """
                File, where :
                - the first data line is the length in meters
                - the second data line is the number of points in the h. plane
                - the third data line is the number of points in the v. plane
                - each data block comes after a START
                - first the horizontal data block, and second the
                      vertical data block
                - each block is a table with axes
                - comments start with #
                """
                # File example:
                # #comment in line 1
                # #comment in line 2
                # Length_in_m
                # #comment in line 4
                # Number of points in horizontal plane :nh
                # #comment in line 6
                # Number of points in vertical plane :nv
                # #comment in line 8
                # START
                #             pos_point1h pos_point2h ... pos_pointnh
                # pos_point1v
                # ...                    horizontal kick_map(nv,nh)
                # pos_pointnv
                # START
                #             pos_point1h pos_point2h ... pos_pointnh
                # pos_point1v
                # ...                    vertical kick_map(nv,nh)
                # pos_pointnv
                # (EOL)

                data_lines = 0     # line not starting with '#'
                header_lines = 0   # line starting with '#'
                block_counter = 0  # START of the h.map, START of the v.map
                for line in f:
                    sline = line.split()
                    if sline[0] == '#':  # line is comment
                        header_lines += 1
                    else:
                        data_lines += 1
                        if data_lines == 1:  # get the element length
                            el_length = float(sline[0])
                        elif data_lines == 2:  # get the number of hor. points
                            h_points = int(sline[0])
                        elif data_lines == 3:  # get the number of ver. points
                            v_points = int(sline[0])
                            # initialize element kicks and table_axes
                            kick_map = numpy.zeros((v_points, h_points))
                            haxis = numpy.zeros(h_points)
                            vaxis = numpy.zeros(v_points)
                        else:
                            # read block of data
                            if sline[0] == "START" or sline[0] == "START\n":
                                block_counter += 1
                                block_lines = 0
                            if block_lines == 1:
                                haxis = sline
                            if block_lines > 1:
                                # minus one due to python index starting at 0
                                # and minus another one due
                                # to the column labels in first line
                                vaxis[block_lines - 2] = float(sline[0])
                                kick_map[block_lines - 2][:] = sline[1:]
                            if block_lines > v_points:
                                block_lines = 0
                                if block_counter == 1:
                                    hkickmap = numpy.copy(kick_map)
                                    table_cols1 = haxis
                                    table_rows1 = vaxis
                                if block_counter == 2:
                                    vkickmap = numpy.copy(kick_map)
                                    table_cols2 = haxis
                                    table_rows2 = vaxis
                                if block_counter > 2:
                                    print('atWarning: only two tables read')
                            block_lines += 1
            # dummy variables not implemented in the reading function
            # but required
            hkickmap1 = 0.0 * numpy.copy(hkickmap)
            vkickmap1 = 0.0 * numpy.copy(vkickmap)

            return el_length, hkickmap, vkickmap, table_cols1, table_rows1, \
                table_cols2, table_rows2, h_points, v_points, \
                hkickmap1, vkickmap1

        def sorted_table(table_in, sorted_index, order_axis):
            # numpy.asfortranarray makes a copy of contiguous memory positions
            table_out = numpy.copy(table_in)
            for i, iis in zip(range(len(sorted_index)), sorted_index):
                if order_axis == 'col':
                    table_out[:, i] = table_in[:, iis]
                if order_axis == 'row':
                    table_out[i, :] = table_in[iis, :]
            table_out2 = numpy.asfortranarray(table_out)
            return table_out2

        # read the input data
        el_length, hkickmap, vkickmap, \
            table_cols1, table_rows1, \
            table_cols2, table_rows2, \
            NumX, NumY, \
            hkickmap1, vkickmap1 \
            = readRadiaFieldMap(Filename_in)

        # set to float
        table_cols1array = numpy.array(table_cols1, dtype='float64')
        table_rows1array = numpy.array(table_rows1, dtype='float64')
        table_cols2array = numpy.array(table_cols2, dtype='float64')
        table_rows2array = numpy.array(table_rows2, dtype='float64')

        # Reorder table_axes
        cols1sorted_index = numpy.argsort(table_cols1array)
        table_cols1array.sort()
        rows1sorted_index = numpy.argsort(table_rows1array)
        table_rows1array.sort()
        cols2sorted_index = numpy.argsort(table_cols2array)
        table_cols2array.sort()
        rows2sorted_index = numpy.argsort(table_rows2array)
        table_rows2array.sort()
        # Reorder kickmap
        hkickmap_a = sorted_table(hkickmap, cols1sorted_index, 'col')
        hkickmap = sorted_table(hkickmap_a, rows1sorted_index, 'row')
        vkickmap_a = sorted_table(vkickmap, cols2sorted_index, 'col')
        vkickmap = sorted_table(vkickmap_a, rows2sorted_index, 'row')
        # Reorder kickmap1
        hkickmap1_a = sorted_table(hkickmap1, cols1sorted_index, 'col')
        hkickmap1 = sorted_table(hkickmap1_a, rows1sorted_index, 'row')
        vkickmap1_a = sorted_table(vkickmap1, cols2sorted_index, 'col')
        vkickmap1 = sorted_table(vkickmap1_a, rows2sorted_index, 'row')

        # Field to kick factors
        Brho = 1e9 * Energy/clight
        factor = 1.0/(Brho**2)
        xkick = factor * hkickmap
        ykick = factor * vkickmap
        # kick1 vars set to zero, not yet implemented
        factor1 = -1.0/(Brho)
        xkick1 = factor1 * hkickmap1
        ykick1 = factor1 * vkickmap1
        xtable = table_cols1array.T
        ytable = table_rows1array.T

        args_dict = {'PassMethod': 'IdTablePass',
                     'Filename_in': Filename_in,
                     'Normalization_energy': Energy,
                     'Nslice': numpy.uint8(Nslice),
                     'Length': el_length,
                     'xkick': xkick,
                     'ykick': ykick,
                     'xkick1': xkick1,
                     'ykick1': ykick1,
                     'xtable': xtable,
                     'ytable': ytable,
                     }
        return args_dict

    def __init__(self, family_name: str, *args, **kwargs):
        _argnames = ['PassMethod',
                     'Filename_in',
                     'Normalization_energy',
                     'Nslice',
                     'Length',
                     'xkick',
                     'ykick',
                     'xkick1',
                     'ykick1',
                     'xtable',
                     'ytable']
        if len(args) < 11:
            # get data from text file
            elemargs = self.from_text_file(*args)
        else:
            # get data from arguments
            elemargs = dict(zip(_argnames, args))
        elemargs.update(kwargs)
        super().__init__(family_name, **elemargs)
# EOF



import numpy as np
from at.constants import clight, e_mass, qe
import matplotlib.pyplot as plt
import at
from at.collective import Wake, LongResonatorElement

# Set up the ring
nturns = 10000
Npart = 10000
Nbunches = 1
current = 0.1   # A

ring = at.load_m('../../../machine_data/esrf.m')
ring.radiation_off()

ring.set_rf_frequency()
freqres = ring.rf_frequency

ring.set_fillpattern(Nbunches)
ring.beam_current = current

fring, _ = at.fast_ring(ring)
# Switch on RF cavity but have all other radiation sources off
fring.radiation_on(cavity_pass='RFCavityPass',
                   dipole_pass=None, quadrupole_pass=None)

# Define the resonator parameters and current
wturns = 50
srange = Wake.build_srange(0., 0.3, 1.0e-5, 1.0e-2,
                           ring.circumference, ring.circumference * wturns)

detuneHz = -5e4
fr = freqres + detuneHz
qfactor = 4500
rshunt = 6e6
bucket_size = clight/freqres


welem = LongResonatorElement('wake', ring, srange, fr,
                             qfactor, rshunt, Nturns=wturns, Nslice=1)
fring.append(welem)

# Define beam monitor
bmon = at.BeamMoments('mon')
fring.append(bmon)

# Particle generation and tracking
sigm = at.sigma_matrix(ring.radiation_on(copy=True))
part = at.beam(Npart, sigm)

fring.track(part, nturns=nturns, refpts=None, in_place=True)

dp_all = bmon.means[4, 0, :]

# Fit the results to obtain the simulated growth rate
width = 1000
step = 600
xr = (2 * np.arange(int(nturns/step)) * step + width) / 2
amp1 = np.array([np.amax(dp_all[ii * step: ii * step + width])
                 for ii in np.arange(int(nturns / step))])
fit1 = np.polyfit(xr, np.log(amp1), 1)
plotrange = np.arange(nturns)


# Define the parameters we need for the analytical value
gamma = ring.gamma
t_rev = 1.0/ring.revolution_frequency
eta = ring.slip_factor
params = ring.radiation_parameters()
fs = params.f_s
intensity = current * t_rev / qe / Nbunches


# Compute the growth rate
Rp = rshunt / (1 + 1j * qfactor * ((freqres + fs) / fr - fr / (freqres + fs)))
Rm = rshunt / (1 + 1j * qfactor * ((freqres - fs) / fr - fr / (freqres - fs)))

numer = e_mass * 2 * gamma * t_rev**2 * 2 * np.pi * fs
denom = -qe * intensity * eta * 2 * np.pi * fr * np.real(Rp - Rm)
tau_RS = numer / denom
gr = tau_RS/t_rev


lab = (r'$\tau_{{fit}}={:.1f}\ \mathrm{{turns}}$' '\n'
       r'$\tau_{{Chao}}={:.1f}\ \mathrm{{turns}}$').format(1/fit1[0], gr)
# Plot all together
fig = plt.figure(figsize=(9, 5))
ax1 = fig.add_subplot(111)
ax1.plot(dp_all,
         label=r'$\Delta f = {:.1f}\ \mathrm{{kHz}}$'.format(detuneHz/1e3),
         color='b')

ax1.plot(plotrange, np.exp(fit1[1])*np.exp(fit1[0]*plotrange),
         color='k', linestyle='solid', marker='None',
         label=lab)
ax1.plot(xr, amp1, linestyle='None', marker='x', color='r')

ax1.legend(bbox_to_anchor=(1.01, 1))
ax1.set_xlabel('Turn')
ax1.set_ylabel('dp/p')
fig.subplots_adjust(right=0.7)
plt.show()



"""Python port of the Accelerator Toolbox"""

from ._version import __version__, __version_tuple__
# Make all functions visible in the at namespace:
from .lattice import *
from .tracking import *
from .physics import *
from .load import *
from .matching import *
from .acceptance import *
from .collective import *



"""Acceptance computation"""
import numpy
from .boundary import GridMode
# noinspection PyProtectedMember
from .boundary import boundary_search
from typing import Optional, Sequence
import multiprocessing
from ..lattice import Lattice, Refpts, frequency_control, AtError


__all__ = ['get_acceptance', 'get_1d_acceptance', 'get_horizontal_acceptance',
           'get_vertical_acceptance', 'get_momentum_acceptance']


@frequency_control
def get_acceptance(
        ring: Lattice, planes, npoints, amplitudes,
        nturns: Optional[int] = 1024,
        refpts: Optional[Refpts] = None,
        dp: Optional[float] = None,
        offset: Sequence[float] = None, bounds=None,
        grid_mode: Optional[GridMode] = GridMode.RADIAL,
        use_mp: Optional[bool] = False,
        verbose: Optional[bool] = True,
        divider: Optional[int] = 2,
        shift_zero: Optional[float] = 1.0e-6,
        start_method: Optional[str] = None,
):
    # noinspection PyUnresolvedReferences
    r"""Computes the acceptance at ``repfts`` observation points

    Parameters:
        ring:           Lattice definition
        planes:         max. dimension 2, Plane(s) to scan for the acceptance.
          Allowed values are: ``'x'``, ``'xp'``, ``'y'``,
          ``'yp'``, ``'dp'``, ``'ct'``
        npoints:        (len(planes),) array: number of points in each
          dimension
        amplitudes:     (len(planes),) array: set the search range:

          * :py:attr:`GridMode.CARTESIAN/RADIAL <.GridMode.RADIAL>`:
            max. amplitude
          * :py:attr:`.GridMode.RECURSIVE`: initial step
        nturns:         Number of turns for the tracking
        refpts:         Observation points. Default: start of the machine
        dp:             static momentum offset
        offset:         initial orbit. Default: closed orbit
        bounds:         defines the tracked range: range=bounds*amplitude.
          It can be use to select quadrants. For example, default values are:

          * :py:attr:`.GridMode.CARTESIAN`: ((-1, 1), (0, 1))
          * :py:attr:`GridMode.RADIAL/RECURSIVE <.GridMode.RADIAL>`: ((0, 1),
            (:math:`\pi`, 0))
        grid_mode:      defines the evaluation grid:

          * :py:attr:`.GridMode.CARTESIAN`: full [:math:`\:x, y\:`] grid
          * :py:attr:`.GridMode.RADIAL`: full [:math:`\:r, \theta\:`] grid
          * :py:attr:`.GridMode.RECURSIVE`: radial recursive search
        use_mp:         Use python multiprocessing (:py:func:`.patpass`,
          default use :py:func:`.lattice_pass`).
        verbose:        Print out some information
        divider:        Value of the divider used in
          :py:attr:`.GridMode.RECURSIVE` boundary search
        shift_zero: Epsilon offset applied on all 6 coordinates
        start_method:   Python multiprocessing start method. The default
          ``None`` uses the python default that is considered safe.
          Available parameters: ``'fork'``, ``'spawn'``, ``'forkserver'``.
          The default for linux is ``'fork'``, the default for MacOS and
          Windows is ``'spawn'``. ``'fork'`` may used for MacOS to speed-up
          the calculation or to solve runtime errors, however  it is
          considered unsafe.

    Returns:
        boundary:   (2,n) array: 2D acceptance
        survived:   (2,n) array: Coordinates of surviving particles
        tracked:    (2,n) array: Coordinates of tracked particles

    In case of multiple refpts, return values are lists of arrays, with one
    array per ref. point.

    Examples:

        >>> bf,sf,gf = ring.get_acceptance(planes, npoints, amplitudes)
        >>> plt.plot(*gf,'.')
        >>> plt.plot(*sf,'.')
        >>> plt.plot(*bf)
        >>> plt.show()

    .. note::

       * When``use_mp=True`` all the available CPUs will be used.
         This behavior can be changed by setting
         ``at.DConstant.patpass_poolsize`` to the desired value
    """
    kwargs = {}
    if start_method is not None:
        kwargs['start_method'] = start_method

    if verbose:
        nproc = multiprocessing.cpu_count()
        print('\n{0} cpu found for acceptance calculation'.format(nproc))
        if use_mp:
            nprocu = nproc
            print('Multi-process acceptance calculation selected...')
            if nproc == 1:
                print('Consider use_mp=False for single core computations')
        else:
            nprocu = 1
            print('Single process acceptance calculation selected...')
            if nproc > 1:
                print('Consider use_mp=True for parallelized computations')
        np = numpy.atleast_1d(npoints)
        na = 2
        if len(np) == 2:
            na = np[1]
        npp = numpy.prod(npoints)
        rpp = 2*numpy.ceil(numpy.log2(np[0]))*numpy.ceil(na/nprocu)
        mpp = npp/nprocu
        if rpp > mpp:
            cond = (grid_mode is GridMode.RADIAL or
                    grid_mode is GridMode.CARTESIAN)
        else:
            cond = grid_mode is GridMode.RECURSIVE
        if rpp > mpp and not cond:
            print('The estimated load for grid mode is {0}'.format(mpp))
            print('The estimated load for recursive mode is {0}'.format(rpp))
            print('{0} or {1} is recommended'.format(GridMode.RADIAL,
                                                     GridMode.CARTESIAN))
        elif rpp < mpp and not cond:
            print('The estimated load for grid mode is {0}'.format(mpp))
            print('The estimated load for recursive mode is {0}'.format(rpp))
            print('{0} is recommended'.format(GridMode.RECURSIVE))

    boundary = []
    survived = []
    grid = []
    if refpts is not None:
        rp = ring.uint32_refpts(refpts)
    else:
        rp = numpy.atleast_1d(refpts)
    if offset is not None:
        try:
            offset = numpy.broadcast_to(offset, (len(rp), 6))
        except ValueError:
            msg = ('offset and refpts have incoherent '
                   'shapes: {0}, {1}'.format(numpy.shape(offset),
                                             numpy.shape(refpts)))
            raise AtError(msg)
    else:
        offset=[None for _ in rp]
    for r, o in zip(rp, offset):
        b, s, g = boundary_search(ring, planes, npoints, amplitudes,
                                  nturns=nturns, obspt=r, dp=dp,
                                  offset=o, bounds=bounds,
                                  grid_mode=grid_mode, use_mp=use_mp,
                                  verbose=verbose, divider=divider,
                                  shift_zero=shift_zero, **kwargs)
        boundary.append(b)
        survived.append(s)
        grid.append(g)
    if len(rp) == 1:
        return boundary[0], survived[0], grid[0]
    else:
        return boundary, survived, grid


def get_1d_acceptance(
        ring: Lattice, plane: str, resolution: float, amplitude: float,
        nturns: Optional[int] = 1024,
        refpts: Optional[Refpts] = None,
        dp: Optional[float] = None,
        offset: Sequence[float] = None,
        grid_mode: Optional[GridMode] = GridMode.RADIAL,
        use_mp: Optional[bool] = False,
        verbose: Optional[bool] = False,
        divider: Optional[int] = 2,
        shift_zero: Optional[float] = 1.0e-6,
        start_method: Optional[str] = None,

):
    r"""Computes the 1D acceptance at ``refpts`` observation points

    See :py:func:`get_acceptance`

    Parameters:
        ring:           Lattice definition
        plane:          Plane to scan for the acceptance.
          Allowed values are: ``'x'``, ``'xp'``, ``'y'``, ``'yp'``, ``'dp'``,
          ``'ct'``
        resolution:     Minimum distance between 2 grid points
        amplitude:      Search range:

          * :py:attr:`GridMode.CARTESIAN/RADIAL <.GridMode.RADIAL>`:
            max. amplitude
          * :py:attr:`.GridMode.RECURSIVE`: initial step
        nturns:         Number of turns for the tracking
        refpts:         Observation points. Default: start of the machine
        dp:             static momentum offset
        offset:         initial orbit. Default: closed orbit
        grid_mode:      defines the evaluation grid:

          * :py:attr:`.GridMode.CARTESIAN`: full [:math:`\:x, y\:`] grid
          * :py:attr:`.GridMode.RADIAL`: full [:math:`\:r, \theta\:`] grid
          * :py:attr:`.GridMode.RECURSIVE`: radial recursive search
        use_mp:         Use python multiprocessing (:py:func:`.patpass`,
          default use :py:func:`.lattice_pass`). In case multi-processing
          is not enabled, ``grid_mode`` is forced to
          :py:attr:`.GridMode.RECURSIVE` (most efficient in single core)
        verbose:        Print out some information
        divider:        Value of the divider used in
          :py:attr:`.GridMode.RECURSIVE` boundary search
        shift_zero: Epsilon offset applied on all 6 coordinates
        start_method:   Python multiprocessing start method. The default
          ``None`` uses the python default that is considered safe.
          Available parameters: ``'fork'``, ``'spawn'``, ``'forkserver'``.
          The default for linux is ``'fork'``, the default for MacOS and
          Windows is ``'spawn'``. ``'fork'`` may used for MacOS to speed-up
          the calculation or to solve runtime errors, however  it is considered
          unsafe.

    Returns:
        boundary:   (len(refpts),2) array: 1D acceptance
        tracked:    (n,) array: Coordinates of tracked particles
        survived:   (n,) array: Coordinates of surviving particles

    In case of multiple ``tracked`` and ``survived`` are lists of arrays,
    with one array per ref. point.

    .. note::

       * When``use_mp=True`` all the available CPUs will be used.
         This behavior can be changed by setting
         ``at.DConstant.patpass_poolsize`` to the desired value
    """
    if not use_mp:
        grid_mode = GridMode.RECURSIVE
    assert len(numpy.atleast_1d(plane)) == 1, \
        '1D acceptance: single plane required'
    assert numpy.isscalar(resolution), '1D acceptance: scalar args required'
    assert numpy.isscalar(amplitude), '1D acceptance: scalar args required'
    npoint = numpy.ceil(amplitude/resolution)
    if grid_mode is not GridMode.RECURSIVE:
        assert npoint > 1, \
            'Grid has only one point: increase amplitude or reduce resolution'
    b, s, g = get_acceptance(ring, plane, npoint, amplitude,
                             nturns=nturns, dp=dp, refpts=refpts,
                             grid_mode=grid_mode, use_mp=use_mp,
                             verbose=verbose, start_method=start_method,
                             divider=divider, shift_zero=shift_zero,
                             offset=offset)
    return numpy.squeeze(b), s, g


def get_horizontal_acceptance(ring: Lattice,
                              resolution: float, amplitude: float,
                              *args, **kwargs):
    r"""Computes the 1D horizontal acceptance at ``refpts`` observation points

    See :py:func:`get_acceptance`

    Parameters:
        ring:           Lattice definition
        resolution:     Minimum distance between 2 grid points
        amplitude:      Search range:

          * :py:attr:`GridMode.CARTESIAN/RADIAL <.GridMode.RADIAL>`:
            max. amplitude
          * :py:attr:`.GridMode.RECURSIVE`: initial step

    Keyword Args:
        nturns:         Number of turns for the tracking
        refpts:         Observation points. Default: start of the machine
        dp:             static momentum offset
        offset:         initial orbit. Default: closed orbit
        grid_mode:      defines the evaluation grid:

          * :py:attr:`.GridMode.CARTESIAN`: full [:math:`\:x, y\:`] grid
          * :py:attr:`.GridMode.RADIAL`: full [:math:`\:r, \theta\:`] grid
          * :py:attr:`.GridMode.RECURSIVE`: radial recursive search
        use_mp:         Use python multiprocessing (:py:func:`.patpass`,
          default use :py:func:`.lattice_pass`). In case multi-processing
          is not enabled, ``grid_mode`` is forced to
          :py:attr:`.GridMode.RECURSIVE` (most efficient in single core)
        verbose:        Print out some information
        divider:        Value of the divider used in
          :py:attr:`.GridMode.RECURSIVE` boundary search
        shift_zero: Epsilon offset applied on all 6 coordinates
        start_method:   Python multiprocessing start method. The default
          ``None`` uses the python default that is considered safe.
          Available parameters: ``'fork'``, ``'spawn'``, ``'forkserver'``.
          The default for linux is ``'fork'``, the default for MacOS and
          Windows is ``'spawn'``. ``'fork'`` may used for MacOS to speed-up
          the calculation or to solve runtime errors, however  it is considered
          unsafe.

    Returns:
        boundary:   (len(refpts),2) array: 1D acceptance
        tracked:    (n,) array: Coordinates of tracked particles
        survived:   (n,) array: Coordinates of surviving particles

    In case of multiple ``tracked`` and ``survived`` are lists of arrays,
    with one array per ref. point.

    .. note::

       * When``use_mp=True`` all the available CPUs will be used.
         This behavior can be changed by setting
         ``at.DConstant.patpass_poolsize`` to the desired value
    """
    return get_1d_acceptance(ring, 'x', resolution, amplitude, *args, **kwargs)


def get_vertical_acceptance(ring: Lattice,
                            resolution: float, amplitude: float,
                            *args, **kwargs):
    r"""Computes the 1D vertical acceptance at refpts observation points

    See :py:func:`get_acceptance`

    Parameters:
        ring:           Lattice definition
        resolution:     Minimum distance between 2 grid points
        amplitude:      Search range:

          * :py:attr:`GridMode.CARTESIAN/RADIAL <.GridMode.RADIAL>`:
            max. amplitude
          * :py:attr:`.GridMode.RECURSIVE`: initial step

    Keyword Args:
        nturns:         Number of turns for the tracking
        refpts:         Observation points. Default: start of the machine
        dp:             static momentum offset
        offset:         initial orbit. Default: closed orbit
        grid_mode:      defines the evaluation grid:

          * :py:attr:`.GridMode.CARTESIAN`: full [:math:`\:x, y\:`] grid
          * :py:attr:`.GridMode.RADIAL`: full [:math:`\:r, \theta\:`] grid
          * :py:attr:`.GridMode.RECURSIVE`: radial recursive search
        use_mp:         Use python multiprocessing (:py:func:`.patpass`,
          default use :py:func:`.lattice_pass`). In case multi-processing
          is not enabled, ``grid_mode`` is forced to
          :py:attr:`.GridMode.RECURSIVE` (most efficient in single core)
        verbose:        Print out some information
        divider:        Value of the divider used in
          :py:attr:`.GridMode.RECURSIVE` boundary search
        shift_zero: Epsilon offset applied on all 6 coordinates
        start_method:   Python multiprocessing start method. The default
          ``None`` uses the python default that is considered safe.
          Available parameters: ``'fork'``, ``'spawn'``, ``'forkserver'``.
          The default for linux is ``'fork'``, the default for MacOS and
          Windows is ``'spawn'``. ``'fork'`` may used for MacOS to speed-up
          the calculation or to solve runtime errors, however  it is considered
          unsafe.

    Returns:
        boundary:   (len(refpts),2) array: 1D acceptance
        tracked:    (n,) array: Coordinates of tracked particles
        survived:   (n,) array: Coordinates of surviving particles

    In case of multiple ``tracked`` and ``survived`` are lists of arrays,
    with one array per ref. point.

    .. note::

       * When``use_mp=True`` all the available CPUs will be used.
         This behavior can be changed by setting
         ``at.DConstant.patpass_poolsize`` to the desired value
    """
    return get_1d_acceptance(ring, 'y', resolution, amplitude, *args, **kwargs)


def get_momentum_acceptance(ring: Lattice,
                            resolution: float, amplitude: float,
                            *args, **kwargs):
    r"""Computes the 1D momentum acceptance at refpts observation points

    See :py:func:`get_acceptance`

    Parameters:
        ring:           Lattice definition
        resolution:     Minimum distance between 2 grid points
        amplitude:      Search range:

          * :py:attr:`GridMode.CARTESIAN/RADIAL <.GridMode.RADIAL>`:
            max. amplitude
          * :py:attr:`.GridMode.RECURSIVE`: initial step

    Keyword Args:
        nturns:         Number of turns for the tracking
        refpts:         Observation points. Default: start of the machine
        dp:             static momentum offset
        offset:         initial orbit. Default: closed orbit
        grid_mode:      defines the evaluation grid:

          * :py:attr:`.GridMode.CARTESIAN`: full [:math:`\:x, y\:`] grid
          * :py:attr:`.GridMode.RADIAL`: full [:math:`\:r, \theta\:`] grid
          * :py:attr:`.GridMode.RECURSIVE`: radial recursive search
        use_mp:         Use python multiprocessing (:py:func:`.patpass`,
          default use :py:func:`.lattice_pass`). In case multi-processing is
          not enabled, ``grid_mode`` is forced to
          :py:attr:`.GridMode.RECURSIVE` (most efficient in single core)
        verbose:        Print out some information
        divider:        Value of the divider used in
          :py:attr:`.GridMode.RECURSIVE` boundary search
        shift_zero: Epsilon offset applied on all 6 coordinates
        start_method:   Python multiprocessing start method. The default
          ``None`` uses the python default that is considered safe.
          Available parameters: ``'fork'``, ``'spawn'``, ``'forkserver'``.
          The default for linux is ``'fork'``, the default for MacOS and
          Windows is ``'spawn'``. ``'fork'`` may used for MacOS to speed-up
          the calculation or to solve runtime errors, however  it is considered
          unsafe.

    Returns:
        boundary:   (len(refpts),2) array: 1D acceptance
        tracked:    (n,) array: Coordinates of tracked particles
        survived:   (n,) array: Coordinates of surviving particles

    In case of multiple ``tracked`` and ``survived`` are lists of arrays,
    with one array per ref. point.

    .. note::

       * When``use_mp=True`` all the available CPUs will be used.
         This behavior can be changed by setting
         ``at.DConstant.patpass_poolsize`` to the desired value
    """
    return get_1d_acceptance(ring, 'dp', resolution, amplitude,
                             *args, **kwargs)


Lattice.get_acceptance = get_acceptance
Lattice.get_horizontal_acceptance = get_horizontal_acceptance
Lattice.get_vertical_acceptance = get_vertical_acceptance
Lattice.get_momentum_acceptance = get_momentum_acceptance



"""
conftest.py is a special pytest file that allows you to share fixtures
between other modules.
"""

import os
import sys

import pytest

if sys.version_info.minor < 9:
    from importlib_resources import files, as_file
else:
    from importlib.resources import files, as_file
import platform
import numpy
import machine_data
import at

try:
    from matlab.engine import connect_matlab, start_matlab, EngineError
except ImportError:
    print("Matlab comparison tests require Matlab Python Engine installed.")
    print("Python will exit.")
    sys.exit()

ROOT_DIR = os.path.realpath(os.path.join(os.path.dirname(__file__), "../.."))


# noinspection PyUnusedLocal
def pytest_report_header(config):
    try:
        a = platform.uname()
        sysinfo = "system: {} {}".format(a[0], a[2])
    except AttributeError:
        sysinfo = ""
    numpyinfo = "numpy version: {}".format(numpy.__version__)
    return [sysinfo, numpyinfo]


@pytest.fixture(scope="session")
def engine():
    try:
        eng = connect_matlab("pytest")
        # Keep the existing Matlab path
    except EngineError:
        eng = start_matlab()
        # Add the local AT path
        eng.addpath(eng.genpath(os.path.join(ROOT_DIR, "atintegrators/")))
        eng.addpath(eng.genpath(os.path.join(ROOT_DIR, "atmat/")))
    # Add the local test_matlab directory
    eng.addpath(os.path.dirname(__file__))
    yield eng
    eng.quit()


def _load_lattice(engine, name, key):
    with as_file(files(machine_data) / name) as path:
        myl = engine.load(str(path))
        pyl = at.load_lattice(path, key=key, keep_all=True)
    return pyl, myl[key], None


# ----------------------- dba lattice -----------------------
@pytest.fixture(scope="session")
def dba(engine):
    return _load_lattice(engine, "dba.mat", "RING")


# ----------------------- hmba lattice -----------------------
@pytest.fixture(scope="session")
def hmba(engine):
    return _load_lattice(engine, "hmba.mat", "RING")


#                hmba lattice with cavities on
@pytest.fixture(scope="session")
def hmba_cav(engine):
    pyl, myl, _ = _load_lattice(engine, "hmba.mat", "RING")
    myl, radindex = engine.pyproxy("atradon", myl, "BendPass", "", nargout=2)
    return pyl.radiation_on(copy=True, dipole_pass=None), myl, radindex


#               hmba lattice with radiation on
@pytest.fixture(scope="session")
def hmba_rad(engine):
    pyl, myl, _ = _load_lattice(engine, "hmba.mat", "RING")
    myl, radindex = engine.pyproxy("atradon", myl, nargout=2)
    return pyl.radiation_on(copy=True), myl, radindex


# -------------------- lattice with errors --------------------
# Takes too much time for usual tests
@pytest.fixture(scope="session")
def err(engine):
    return _load_lattice(engine, "err.mat", "RING")
